def initialise_kernel(config_kernel, output_folder, task_id):
    if config_kernel.type == 'montepython':
        from prospect.kernels.montepython_kernel import MontePythonKernel
        kernel = MontePythonKernel(config_kernel, task_id, output_folder=output_folder)
    elif config_kernel.type == 'cobaya':
        from prospect.kernels.cobaya_kernel import CobayaKernel
        kernel = CobayaKernel(config_kernel, task_id, output_folder=output_folder)
    elif config_kernel.type == 'analytical':
        from prospect.kernels.analytical_kernel import AnalyticalKernel
        kernel = AnalyticalKernel(config_kernel, task_id, output_folder=output_folder)
    else:
        raise ValueError('You have specified a non-existing kernel type.')
    return kernel



import heapq
import os
import pickle
import time
import traceback
from collections import defaultdict
from datetime import datetime
from typing import Type
from dataclasses import dataclass
from prospect.input import Configuration
from prospect.tasks.base_task import BaseTask
from prospect.tasks.base_analyse_task import BaseAnalyseTask
from prospect.tasks.analyse_profile_task import AnalyseProfileTask
from prospect.tasks.analyse_global_optimisation_task import AnalyseGlobalOptimisationTask
from prospect.tasks.initialise import initialise_tasks
from multiprocessing import Condition

@dataclass
class TasksState:
    queued: list[Type[BaseTask]]
    # All task dicts are indexed by the task ids 
    unready: dict[int, Type[BaseTask]]
    ongoing: dict[int, Type[BaseTask]]
    done:    dict[int, Type[BaseTask]]
    failed:  dict[int, Type[BaseTask]]
    dependencies: defaultdict[int, list[int]]

class Scheduler:
    def __init__(self, config: Configuration, state: TasksState = None) -> None:
        if state:
            # Resume from snapshot
            self.tasks = state
            # Required to restart all tasks that were ongoing
            for task in self.tasks.ongoing.values():
                self.push_task(task)
            self.tasks.ongoing = {}
        else:
            # Start from config file
            self.tasks = TasksState([], {}, {}, {}, {}, defaultdict(list))
            new_tasks = initialise_tasks(config)
            for task in new_tasks:
                self.push_task(task)
        self.config = config
        self.condition = Condition()
        self.start_time = time.time()
        self.start_date = datetime.now()
        self.next_dump_time = time.time() + config.io.snapshot_interval
        self.sleep_iteration = 0

    def delegate(self, executor) -> None:
        # Executor is a context managed instance of either mpi4py.futures.MPICommExecutor 
        # or concurrent.futures.ThreadPoolExecutor
        while True:
            if not self.tasks.queued:
                if not self.tasks.ongoing:
                    if not self.tasks.unready:
                        break
                    elif self.tasks.failed:
                        print(f"Exiting PROSPECT with one or more failed tasks.")
                        break

                with self.condition:
                    self.condition.wait()
                continue

            ready_task = self.pop_task()
            task_requirements = [self.tasks.done[idx] for idx in ready_task.required_task_ids]
            executor.submit(ready_task.run_return_self, task_requirements).add_done_callback(self.finalize_task)

    def finalize(self, executor) -> None:
        # Things to do before shutting down
        if self.config.io.write:
            if self.config.run.jobtype == 'profile':
                analysis_task = self.get_profile_analysis()
                analysis_task.run([task for task in self.tasks.done.values() if task.id in analysis_task.required_task_ids])
            elif self.config.run.jobtype == 'global_optimisation':
                analysis_task = self.get_global_optimisation_analysis()
                analysis_task.run([task for task in self.tasks.done.values() if task.id in analysis_task.required_task_ids])
            self.dump_snapshot()
            self.status_update()

    def push_task(self, task: Type[BaseTask]) -> None:
        # Pushes task and sets dependencies
        task.time_emit = datetime.now()
        if self.is_task_ready(task):
            # Works since BaseTasks are ordered by their priority
            heapq.heappush(self.tasks.queued, task)
        else:
            self.tasks.unready[task.id] = task
        for req in task.required_task_ids:
            self.tasks.dependencies[req].append(task.id)

    def pop_task(self) -> Type[BaseTask]:
        # Pops first task from task queue and moves to ongoing tasks
        ready_task = heapq.heappop(self.tasks.queued)
        ready_task.time_launch = datetime.now()
        self.tasks.ongoing[ready_task.id] = ready_task
        return ready_task

    def is_task_ready(self, task: Type[BaseTask]) -> bool:
        for req in task.required_task_ids:
            if req not in self.tasks.done:
                return False 
        return True
    
    def fail_task(self, task: Type[BaseTask]) -> None:
        print("\n-------------------------------------------------------------------")
        print(f"WARNING! TASK {task.id} RAISED THE FOLLOWING EXCEPTION:\n")
        print(task.error)
        print(f"Moving the failed task to TasksState.failed and continuing.")
        print("-------------------------------------------------------------------\n")
        if hasattr(task, "mcmc.kernel"):
            del task.mcmc.kernel
        del self.tasks.ongoing[task.id]
        self.tasks.failed[task.id] = task
        with self.condition:
            self.condition.notify_all()

    def finalize_task(self, future) -> None:
        finished_task = future.result() # BaseTask.run_return_self returns the task that was run 

        # Handle tasks that raised an exception
        if not finished_task.success:
            self.fail_task(finished_task)
            return

        # Continue, assuming task finished succesfully
        finished_task.time_finish = datetime.now()
        del self.tasks.ongoing[finished_task.id]
        self.tasks.done[finished_task.id] = finished_task

        # Add tasks emitted by finished_task
        for new_task in finished_task.emit_tasks():
            new_task.emitted_by = finished_task.id
            self.push_task(new_task)

        # Update dependencies
        for dependency_id in self.tasks.dependencies[finished_task.id]:
            if dependency_id in self.tasks.unready:
                # Check if dependency is now ready
                if self.is_task_ready(self.tasks.unready[dependency_id]):
                    new_task = self.tasks.unready.pop(dependency_id)
                    heapq.heappush(self.tasks.queued, new_task)

        with self.condition:
            self.condition.notify_all()

        if self.config.io.write:
            if time.time() > self.next_dump_time:
                if finished_task.type != 'AnalyseProfileTask' and finished_task.type != 'AnalyseGlobalOptimisationTask':
                    analyse_task = self.get_analysis_task()
                    if analyse_task is not None:
                        self.push_task(analyse_task)
                self.dump_snapshot()
                self.status_update()

                self.next_dump_time = time.time() + self.config.io.snapshot_interval
    
    def dump_snapshot(self) -> None:
        print("Dumping snapshot...")
        for name, data in [("config", self.config), ("state", self.tasks)]:
            backup_filename = os.path.join(self.config.io.dir, f"{name}_backup.pkl")
            with open(backup_filename, 'wb') as file:
                # Write to backup to minimize risk of crash while saving
                pickle.dump(data, file) 
                os.system(f'mv {backup_filename} {os.path.join(self.config.io.dir, f"{name}.pkl")}')

    def status_update(self) -> None:
        """Writes status of all tasks in status.txt"""
        print("Writing status update...")
        pad_id = 4
        pad_tasktype = 20
        pad_time = 30
        pad_emit = pad_id + 2
        
        def get_write_line(task: Type[BaseTask]) -> str:
            emit_id = task.emitted_by if task.emitted_by is not None else '-'
            out = f"\n{task.id:<{pad_id}} {emit_id:<{pad_emit}} {task.type:<{pad_tasktype}} \t {task.time_emit.ctime():<{pad_time}}"
            if task.time_launch is not None:
                out += f" {task.time_launch.ctime():<{pad_time}}"
            if task.time_finish is not None:
                out += f" {task.time_finish.ctime():<{pad_time}}"
            return out

        with open(os.path.join(self.config.io.dir, "status.txt"), 'w') as status_file:
            header = f"\n{'id':<{pad_id}} {'from':<{pad_emit}} {'task type':<{pad_tasktype}} \t {'Time when emitted':<{pad_time}} {'Time when started':<{pad_time}} {'Time when finished':<{pad_time}}"
            status_file.write(f"Status of job '{self.config.io.jobname}'\nJobtype: {self.config.run.jobtype}\nStarted: {self.start_date}\nLast updated: {datetime.now()}")
            status_file.write("\n\n=== FAILED ==============================================")
            status_file.write(header)
            for task_id, task_failed in self.tasks.failed.items():
                status_file.write(get_write_line(task_failed))
            status_file.write("\n\n=== IN PROGRESS ==========================================")
            status_file.write(header)
            for task_id, task_ongoing in self.tasks.ongoing.items():
                status_file.write(get_write_line(task_ongoing))
            status_file.write("\n\n=== QUEUED, READY =========================================")
            status_file.write(header)
            for task in self.tasks.queued:
                status_file.write(get_write_line(task))
            status_file.write("\n\n=== QUEUED, NOT READY ====================================")
            status_file.write(header)
            for task_id, task_unready in self.tasks.unready.items():
                status_file.write(get_write_line(task_unready))
            status_file.write("\n\n=== DONE ================================================")
            status_file.write(header)
            for task_id, task_done in self.tasks.done.items():
                status_file.write(get_write_line(task_done))
            status_file.write("\n")
    
    def get_analysis_task(self) -> Type[BaseAnalyseTask]:
        required_tasks = [id for id, task in self.tasks.done.items() if task.type == 'OptimiseTask' or task.type == 'InitialiseOptimiserTask']
        if self.config.run.jobtype == 'profile':
            if 'AnalyseProfileTask' not in [queued_task.type for queued_task in self.tasks.queued]:
                if 'AnalyseProfileTask' not in [ongoing_task.type for ongoing_task in self.tasks.ongoing.values()]:
                    return AnalyseProfileTask(self.config, required_tasks)
        elif self.config.run.jobtype == 'global_optimisation':
            if 'AnalyseGlobalOptimisationTask' not in [queued_task.type for queued_task in self.tasks.queued]:
                if 'AnalyseGlobalOptimisationTask' not in [ongoing_task.type for ongoing_task in self.tasks.ongoing.values()]:
                    return AnalyseGlobalOptimisationTask(self.config, required_tasks)

class SerialContext:
    class MockFuture:
        def __init__(self, finished_task: Type[BaseTask]) -> None:
            self.finished_task = finished_task

        def add_done_callback(self, finalize_func) -> None:
            finalize_func(self)

        def result(self) -> Type[BaseTask]:
            return self.finished_task

    def __enter__(self) -> None:
        return self

    def __exit__(self, exc_type, exc_value, tb) -> bool:
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        return True

    def submit(self, run_func, *args) -> MockFuture:
        finished_task = run_func(*args)
        return self.MockFuture(finished_task)




from typing import Type
from prospect.tasks.base_task import BaseTask
from prospect.tasks.mcmc_task import initialise_mcmc_tasks
from prospect.tasks.initialise_profile_task import initialise_profile_tasks

def initialise_tasks(config) -> list[Type[BaseTask]]:
    if config.run.jobtype == 'mcmc':
        new_tasks = initialise_mcmc_tasks(config, mcmc_id=0)
    elif config.run.jobtype == 'profile' or config.run.jobtype == 'global_optimisation':
        new_tasks = initialise_profile_tasks(config)
    else:
        raise ValueError(f'No tasks to start at initialization with jobtype {[config.run.jobtype]}.')
    return new_tasks


import contextlib
import inspect
import os
from abc import ABC
from dataclasses import dataclass, field
from importlib import import_module
from types import UnionType
from typing import Any, get_args, get_origin, Union
import numpy as np 

class Configuration:
    def __init__(self, config_yaml):
        """config_yaml is the dictionary read from the input .yaml file."""
        print("Reading input.")
        # When new arguments are implemented, add them here
        modules = {
            'run':    import_module('prospect.run'), 
            'io':     import_module('prospect.io'),
            'kernel': import_module('prospect.kernels.base_kernel')
        }
        if 'run' in config_yaml and 'jobtype' in config_yaml['run']:
            if config_yaml['run']['jobtype'] == 'mcmc':
                modules['mcmc'] = import_module('prospect.mcmc')
            elif config_yaml['run']['jobtype'] == 'profile':
                modules['profile'] = import_module('prospect.profile')
            elif config_yaml['run']['jobtype'] == 'global_optimisation':
                # Global optimisation uses profiling inputs for optimisation
                modules['profile'] = import_module('prospect.profile')
            else:
                raise ValueError("The given value of 'jobtype' is not recognised. Choose either 'mcmc' or 'profile'.")
        else:
            raise ValueError("You must give 'jobtype' as an input under the 'run' category.")

        str_eval_arguments = [('profile', 'values')]
        for arg_module, arg_name in str_eval_arguments:
            if arg_module in config_yaml:
                if arg_name in config_yaml[arg_module]:
                    if type(config_yaml[arg_module][arg_name]) == str:
                        config_yaml[arg_module][arg_name] = safe_eval(config_yaml[arg_module][arg_name])

        self.set_defaults_iterative(modules, config_yaml)
        self.validate_parameters(modules, config_yaml)
        self.set_output_dir(config_yaml['io'])

        # Set parameters as attributes
        self.config_dict = config_yaml
        for module_name, module in modules.items():
            setattr(self, module_name, module.Arguments(**config_yaml[module_name]))
        
        print("Succeeded in setting defaults and validating inputs.")
        print("------------------------------")
    

    def set_defaults_iterative(self, modules: list, config_yaml) -> None:
        num_read = 1
        while num_read > 0:
            num_read = 0
            failed_args = []

            # Set default values
            for module_name, module in modules.items():
                print(f"Setting defaults for module '{module_name}'.")
                
                for arg_name, input_arg in self.arguments_iterator(module.Arguments):
                    try:
                        if not arg_name in config_yaml[module_name]:
                            config_yaml[module_name][arg_name] = input_arg().get_default(config_yaml)
                            print(f"Argument {arg_name} not found in input, set to default value {config_yaml[module_name][arg_name]}.")
                            num_read += 1
                    except Exception as e:
                        print(f"Failed setting parameter '{arg_name}'. Exception message:\n{e}")
                        failed_args.append(arg_name)
            if num_read == 0:
                if not failed_args:
                    print(f"End of iterative default setting, continuing to validation.")
                else:
                    raise ValueError(f"Could not set default value for inputs {failed_args}.")
            else:
                print(f"Set {num_read} default parameter(s), reiterating.")

    def arguments_iterator(self, arguments) -> list[str, Any]:
        # arguments is an instance of an Arguments class defined in the particular module
        out = inspect.getmembers(arguments, inspect.isclass)
        for idx, (name, arg) in enumerate(out):
            if name == '__class__':
                out.pop(idx)
        return out

    def validate_parameters(self, modules, config_yaml) -> None:
        for module_name, module in modules.items():
            print(f"Validating input from {module_name}.")
            for arg_name, input_arg in self.arguments_iterator(module.Arguments):
                self.validate_generic(input_arg, config_yaml[module_name][arg_name])
                input_arg().validate(config_yaml[module_name])

    def validate_generic(self, arg, arg_val: Any) -> None:
        if arg.val_type is not None:
            if get_origin(arg.val_type) in [Union, UnionType]:
                if not type(arg_val) in get_args(arg.val_type):
                    raise ValueError(f"Input '{arg().name}' was given type value of {type(arg_val)} which is not one of the allowed types: {get_args(arg.val_type)}.")
            else:
                if not type(arg_val) == arg.val_type:
                    raise ValueError(f"Input '{arg().name}' was given type value of {type(arg_val)}, but only type {arg.val_type} is allowed.")
        if arg.allowed_values is not None:
            if not arg_val in arg.allowed_values:
                raise ValueError(f"Input '{arg().name}' was given value {arg_val} which not in the list of allowed values: {arg.allowed_values}")

    def set_output_dir(self, config_io) -> None:
        # Sets output dir depending on whether to overwrite
        if config_io['write']:
            if not config_io['overwrite_dir']:
                if os.path.isdir(config_io['dir']):
                    output_idx = 0
                    while True:
                        if os.path.isdir(f"{config_io['dir']}_{output_idx}"):
                            output_idx += 1
                        else:
                            break
                    config_io['dir'] = f"{config_io['dir']}_{output_idx}"

def safe_eval(code):
    @contextlib.contextmanager
    def disable_imports():
        import builtins
        __import__ = builtins.__import__
        builtins.__import__ = None
        try:
            yield
        finally:
            builtins.__import__ = __import__
    scope = {'np': np, 'numpy': np}
    with disable_imports():
        try:
            return eval(code, scope)
        except Exception:
            raise ValueError(f'Illegal input {code!r}. Note that you may only use the standard and numpy library in Python-evaluted statements in the .yaml file.')

@dataclass
class InputArgument(ABC):
    # If set, will check that the correct type is given
    val_type: Any = field(default=None) 

    # If not specified, all values are allowed
    allowed_values: list = field(default=None)

    # Specify default values as functions of other arguments
    def get_default(self, config_yaml: dict[str, Any]):
        # config_yaml is the entire input dictionary
        raise ValueError(f"The input '{self.name}' has no default and must be specified.")
        
    def validate(self, config: dict[str, Any]) -> None:
        # config is the config_yaml[module] input dict, after settings default values, where module is the module that this argument is defined in 
        pass

    @property
    def name(self) -> str:
        return self.__class__.__name__



import os
import pickle
import yaml
import numpy as np 
import scipy as sc
from prospect.kernels.base_kernel import BaseKernel

class AnalyticalKernel(BaseKernel):
    # Analytical likelihoods for development, testing and benchmarking purposes
    def initialise(self, config_kernel, output_folder):
        # Instead, I should make the param-file Python-readable to easily make large dims etc.
        self.config = yaml.full_load(open(os.path.join(os.getcwd(), config_kernel.param), 'r'))

        if self.config['function'] == 'gaussian':
            self._loglkl = self.Gaussian
        elif self.config['function'] == 'random_gaussian':
            self._loglkl = self.Gaussian
            self.set_parameter_dict = self.set_parameter_dict_random

            settings_file = os.path.join(output_folder, f'analytical/random_gauss.pkl')
            if not os.path.isfile(settings_file):
                if 'load' in self.config:
                    # Load from user specified setting
                    print(f"Loading random Gaussian kernel from {self.config['load']}...")
                    with open(self.config['load'], 'rb') as file:
                        kernel_dict = pickle.load(file)
                else:
                    # Generate random multivariate Gaussian parameters
                    print("Generating new random Gaussian kernel...")
                    self.config['means'] = {}
                    for idx_param in range(1, self.config['dimension']+1):
                        self.config['means'][f'x{idx_param}'] = np.random.rand()
                    covmat_diag = np.diag(np.random.rand(self.config['dimension']))
                    covmat_offdiag = np.random.rand(self.config['dimension'], self.config['dimension'])
                    self.config['covmat'] = self.config['std_scale']*(covmat_diag + covmat_offdiag*self.config['off_diag_factor'])
                    kernel_dict = {
                        'means': self.config['means'],
                        'covmat': self.config['covmat'],
                        'dimension': self.config['dimension']
                    }
                # Save to disk
                with open(settings_file, 'wb') as file:
                    print(f"Saving new random Gaussian kernel to {settings_file}...")
                    pickle.dump(kernel_dict, file)
            else:
                # Load the previously generated random kernel 
                with open(settings_file, 'rb') as file:
                    kernel_dict = pickle.load(file)
            self.config['means'] = kernel_dict['means']
            self.config['covmat'] = kernel_dict['covmat']
            self.config['dimension'] = kernel_dict['dimension']
        else:
            raise ValueError('No analytical kernel with the desired function type.')
        self.dimension = self.config['dimension']
        self.scipy_profile = {}
        self.config_kernel = config_kernel
        self.output_folder = output_folder
    
    def Gaussian(self, position):
        residual = []
        for param_name, param in self.param['varying'].items():
            residual.append(position[param_name][0] - self.config['means'][param_name])
        for param_name, param in self.param['fixed'].items():
            residual.append(param['fixed_value'] - self.config['means'][param_name])
        loglkl = 0.5*np.dot(residual, np.matmul(np.linalg.inv(self.config['covmat']), residual))
        return loglkl
    
    def set_parameter_dict(self):
        for param_name, prior in self.config['param_dict'].items():
            self.param['varying'][param_name] = {'range': prior}
    
    def _loglkl(self, position):
        raise NotImplementedError("Method '_loglkl' of AnalyticalKernel must be set on initialisation!")
    
    def logprior(self, position):
        return self.log_uniform_prior(position)

    def get_initial_position(self, config_initial_position=None):
        if config_initial_position is not None:
            raise KeyError('The analytical kernel does not currently take an initial position as argument.')
        else:
            # Default
            out = {param: [0.] for param in self.param['varying'].keys()}
            for param_name, param_dict in self.param['fixed']:
                out[param_name] = [param_dict['fixed_value']]
            return out

    def get_default_initial_position(self):
        out = {param: [0.] for param in self.param['varying'].keys()}
        for param_name, param_dict in self.param['fixed']:
            out[param_name] = [param_dict['fixed_value']]
        return out

    def read_initial_position(self, config_initial_position):
        # Input should be formatted as something that can be cast to a numpy array 
        return np.array(config_initial_position)

    def get_default_covmat(self):
        # Make a diagonal covmat with 1/20 of the prior widths as stds in each parameter
        stds = []
        for param_name, param in self.param['varying'].items():
            stds.append(param['range'][1] - param['range'][0])
        return 0.005*np.diag(stds)

    def read_covmat(self, config_covmat):
        # Input should be formatted as something that can be cast to a numpy array 
        return np.array(config_covmat)

    def set_parameter_dict_random(self):
        # Overloads the set_parameter_dict from BaseKernel
        for idx_dim in range(1,self.config['dimension']+1):
            self.param['varying'][f'x{idx_dim}'] = {
                'range': [-2.5, 2.5]
            }
    
    def get_scipy_profile(self, parameter, fixed_val):
        if fixed_val not in self.scipy_profile:
            assert parameter not in self.param['varying']
            self.param['fixed'][parameter]['fixed_value'] = fixed_val
            loglkl_scipy = sc.optimize.minimize(self.wrapped_loglkl, np.zeros(len(self.param['varying']))).fun
            self.scipy_profile[fixed_val] = loglkl_scipy
        return self.scipy_profile[fixed_val]
    
    def wrapped_loglkl(self, array_of_positions: np.ndarray):
        position = {}
        for idx, param_name in enumerate(list(self.param['varying'].keys())):
            position[param_name] = [array_of_positions[idx]]
        return self.loglkl(position)


from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
import os 
import time
from typing import Type
import matplotlib
import matplotlib.pyplot as plt 
import numpy as np 
from scipy.interpolate import interp1d
from scipy import stats
from prospect.input import Configuration
from prospect.kernels.initialisation import initialise_kernel
from prospect.tasks.base_task import BaseTask
from prospect.tasks.optimise_task import OptimiseTask

# Switch matplotlib backend to one that doesn't pop up figures
matplotlib.use('Agg')

def argmax(iterable):
    """From https://stackoverflow.com/questions/16945518/finding-the-index-of-the-value-which-is-the-min-or-max-in-python"""
    return max(enumerate(iterable), key=lambda x: x[1])[0]

class BaseAnalyseTask(BaseTask, ABC):
    priority = 75.0

    # Figure settings
    cm = 1/2.54  # centimeters in inches
    fig = {
        'width': 8.6*cm, # PRL figure width
        'height': 2,
        'fontsize': 11/1.1,
        'rep_colors': ['b', 'r', 'g', 'c', 'y', 'm']*50,
        'rep_ms': [4.5, 4.0, 3.5, 3, 2.5, 2., 1.5, 1., 0.5]*50,
        'interval_styles': ['--', ':']*5
    }

    def __init__(self, config: Configuration, required_task_ids: list[int]):
        super().__init__(config, required_task_ids)
        self.dir = os.path.join(config.io.dir, config.run.jobtype)
        if not os.path.isdir(self.dir):
            os.makedirs(self.dir)


    @abstractmethod
    def run(self, tasks: list[OptimiseTask]): 
        pass

    def emit_tasks(self) -> list[Type[BaseTask]]:
        # Never emit anything; the OptimiseTasks manage convergence themselves.
        return []



from typing import Type
import numpy as np 
from prospect.input import Configuration
from prospect.tasks.base_task import BaseTask
from prospect.kernels.initialisation import initialise_kernel
from prospect.optimiser import initialise_optimiser

class OptimiseTask(BaseTask):
    priority = 25.0

    def __init__(self, config: Configuration, optimise_task_settings: dict):
        super().__init__(config)
        self.optimise_settings = optimise_task_settings
        self.converged = False

    def run(self, _):
        kernel = initialise_kernel(self.config.kernel, self.config.io.dir, self.id)
        if 'fixed_param_val' in self.optimise_settings:
            kernel.set_fixed_parameters({self.config.profile.parameter: self.optimise_settings['fixed_param_val']})
        self.optimiser = initialise_optimiser(self.config, kernel, self.optimise_settings)
        self.optimiser.optimise()
        self.optimiser.set_bestfit()
        if self.config.run.jobtype == 'profile':
            print(f"OptimiseTask id={self.id}, iteration {self.optimise_settings['iteration_number']} at {self.config.profile.parameter}={self.optimise_settings['fixed_param_val']:.4e} finished:")
        else:
            print(f"OptimiseTask id={self.id}, iteration {self.optimise_settings['iteration_number']} finished:")
        print(f"\tBestfit loglkl={self.optimiser.bestfit['loglkl']:.10} | Acceptance rate={self.optimiser.bestfit['acceptance_rate']:.4} | Num_evals={np.sum(self.optimiser.mcmc.chain.mults)} | Temp={self.optimise_settings['temperature']:.4e} | Step size={self.optimise_settings['step_size']:.4e}")

    def finalize(self):
        del self.optimiser.mcmc.kernel
        del self.optimiser.kernel

    def emit_tasks(self) -> list[Type[BaseTask]]:
        loglkl_improvement = self.optimise_settings['current_best_loglkl'] - self.optimiser.bestfit['loglkl']
        if self.config.profile.chi2_tolerance is not None and self.config.profile.chi2_tolerance != 0.0:
            if loglkl_improvement < 0.:
                print(f"Best chi2 didn't improve since last iteration ({self.optimise_settings['current_best_loglkl']:.10}). Emitting next OptimiseTask.")
            elif self.optimise_settings['current_best_loglkl'] == np.inf:
                # First iteration, continue
                pass
            # New convergence criterion: All loglkls appearing in chain must be < tolerance
            # Perhaps a better conv crit is not to compare bestfits but just the loglkl of the last points
            # or alternatively the Euclidean distance between last points (but this is difficult to normalize...)
            #elif np.any(2*np.array(self.optimiser.mcmc.chain.loglkls) > self.config.profile.chi2_tolerance):
            elif 2*loglkl_improvement > self.config.profile.chi2_tolerance:
                print(f"Best chi2 improved by {2*loglkl_improvement:.4e} since last iteration.\nEmitting new OptimiseTask since some loglkls deviated more than tolerance chi2_tolerance={self.config.profile.chi2_tolerance}.")
            elif self.optimiser.bestfit['accepted_steps'] > 1:
                print(f"Best chi2 improved by {2*loglkl_improvement:.4e} since last iteration.\nFinishing optimisation since all loglkls deviated less than tolerance chi2_tolerance={self.config.profile.chi2_tolerance}.")
                self.converged = True
                return []
        new_settings = self.optimiser.get_next_iteration_settings()
        new_settings['iteration_number'] = self.optimise_settings['iteration_number'] + 1
        new_settings['repetition_number'] = self.optimise_settings['repetition_number']
        return [OptimiseTask(self.config, new_settings)]



from collections import defaultdict
from copy import deepcopy
import os 
import time
from typing import Type
import matplotlib
import matplotlib.pyplot as plt 
import numpy as np 
from scipy.interpolate import interp1d
from scipy import stats
from prospect.input import Configuration
from prospect.kernels.initialisation import initialise_kernel
from prospect.tasks.base_analyse_task import BaseAnalyseTask
from prospect.tasks.optimise_task import OptimiseTask

# Switch matplotlib backend to one that doesn't pop up figures
matplotlib.use('Agg')

def argmax(iterable):
    """From https://stackoverflow.com/questions/16945518/finding-the-index-of-the-value-which-is-the-min-or-max-in-python"""
    return max(enumerate(iterable), key=lambda x: x[1])[0]

class AnalyseProfileTask(BaseAnalyseTask):
    def run(self, tasks: list[OptimiseTask]): 
        if not tasks:
            print(f"No tasks to analyse. Ending AnalyseProfileTask.")
            return
        print("Running profile analysis script...")
        kernel = initialise_kernel(self.config.kernel, self.config.io.dir, self.id)
        optimise_tasks, initialise_tasks = [], []
        param_vals = []
        for task in tasks:
            if task.type == 'OptimiseTask':
                if task.optimise_settings['fixed_param_val'] not in param_vals:
                    param_vals.append(task.optimise_settings['fixed_param_val'])
                optimise_tasks.append(task)
            elif task.type == 'InitialiseOptimiserTask':
                initialise_tasks.append(task)

        param_vals = np.sort(param_vals)
        old_param_vals = deepcopy(self.config.profile.values)
        self.config.profile.values = param_vals

        self.bestfits = self.sort_tasks(optimise_tasks, initialise_tasks)
        profile = self.compute_profile(self.bestfits)
        self.write_results(self.bestfits, kernel)
        self.write_profile_status(optimise_tasks, initialise_tasks)

        intervals = self.compute_intervals_neyman(profile)
        self.write_intervals(profile, intervals, kernel)

        if self.config.profile.plot_profile:
            tic = time.perf_counter()
            self.plot_profile(profile, intervals, self.bestfits, kernel)
            print(f"Plotted profile in {time.perf_counter() - tic:.5} s")
        if self.config.profile.plot_schedule:
            tic = time.perf_counter()
            self.plot_schedule(optimise_tasks)
            print(f"Plotted schedule in {time.perf_counter() - tic:.5} s")
        self.config.profile.values = old_param_vals
    
    def write_profile_status(self, optimise_tasks, initialise_tasks):
        status_file = os.path.join(self.dir, f'{self.config.profile.parameter}_status.txt')

        print(f"Writing current profile status to {status_file}")
        with open(status_file,"w") as file:
            # Write header
            file.write(f"{self.config.profile.parameter:14.10} ")
            file.write(f"\t\t | rep. no.  current iter \t current loglkl \t best iter \t best loglkl \t converged? ")
            
            for fixed_val in self.config.profile.values:
                file.write(f"\n{fixed_val:14.10} ")
                for idx_rep in range(self.config.profile.repetitions):
                    rep_tasks = [task for task in optimise_tasks if task.optimise_settings['fixed_param_val'] == fixed_val and task.optimise_settings['repetition_number'] == idx_rep]
                    if rep_tasks:
                        current_task = rep_tasks[argmax([task.optimise_settings['iteration_number'] for task in rep_tasks])]
                        current_iter = current_task.optimiser.bestfit['iteration_number']
                        current_loglkl = current_task.optimiser.bestfit['loglkl']

                        best_task = rep_tasks[argmax([-task.optimiser.bestfit['loglkl'] for task in rep_tasks])]
                        best_iter = best_task.optimiser.bestfit['iteration_number']
                        best_loglkl = best_task.optimiser.bestfit['loglkl']
                        pad_current = 13
                        pad_best = 9
                        write_string = f"| rep {idx_rep+1} \t {current_iter:<{pad_current}} \t {current_loglkl:.10e} \t {best_iter:<{pad_best}} \t {best_loglkl:.10e} \t {current_task.converged} "
                        if idx_rep == 0:
                            file.write(f"\t {write_string}")
                        else:
                            file.write(f"\n\t\t\t\t {write_string}")
    
    def sort_tasks(self, optimise_tasks, initialise_tasks):
        # Order the tasks
        bestfits = {}
        for fixed_value in self.config.profile.values:
            bestfits[fixed_value] = {}
            for idx_rep in range(self.config.profile.repetitions):
                current_bestfit = {'loglkl': np.inf}
                for task in optimise_tasks:
                    if task.optimise_settings['fixed_param_val'] == fixed_value:
                        if task.optimise_settings['repetition_number'] == idx_rep:
                            if task.optimiser.bestfit['loglkl'] < current_bestfit['loglkl']:
                                current_bestfit = task.optimiser.bestfit 
                bestfits[fixed_value][idx_rep] = current_bestfit # store the best in each rep
            
            # Find the best rep for this fixed value
            best_rep = list(bestfits[fixed_value].keys())[np.argmin([rep['loglkl'] for rep in bestfits[fixed_value].values()])]
            if np.any([rep['loglkl'] < np.inf for rep in bestfits[fixed_value].values()]):
                avg_loglkl = np.mean([rep['loglkl'] for rep in bestfits[fixed_value].values() if rep['loglkl'] < np.inf])
            else:
                avg_loglkl = np.inf
            bestfits[fixed_value]['best_rep'] = best_rep
            bestfits[fixed_value]['avg_loglkl'] = avg_loglkl

            for initialise_task in initialise_tasks:
                if initialise_task.fixed_param_val == fixed_value:
                    # All reps are all initialised to the same point
                    bestfits[fixed_value]['initial'] = initialise_task.initial_bestfit
                    bestfits[fixed_value]['initial_loglkl'] = initialise_task.initial_loglkl
                    bestfits[fixed_value]['covmat'] = initialise_task.initial_covmat
                    break
        return bestfits
    
    def compute_profile(self, bestfit_dict):
        """
            Default analysis method is to fit the best point of the profile and its two 
            neighbouring points to a parabola and take the top point of the parabola 
            as the bestfit chi2
        
        """

        # order points in terms of profile parameter
        param_vals, loglkls = [], []
        for param_val, bestfit in bestfit_dict.items():
            param_vals.append(param_val)
            loglkl = bestfit[bestfit['best_rep']]['loglkl']
            if loglkl < np.inf:
                loglkls.append(loglkl)
            else:
                # No optimisations finished, take the initial values
                loglkls.append(bestfit['initial_loglkl'])
        ordering = np.argsort(param_vals)
        param_vals = np.array(param_vals)[ordering]
        loglkls = np.array(loglkls)[ordering]

        best_index = np.argmin(loglkls)
        if best_index == 0 or best_index == len(loglkls) - 1:
            # best point is on one of the edges of the profile
            global_best_param_val = param_vals[best_index]
            global_best_loglkl = loglkls[best_index]
        else:
            # find the three parabola points to get Delta chi2
            x = [param_vals[idx] for idx in [best_index-1, best_index, best_index+1]]
            f = [loglkls[idx] for idx in [best_index-1, best_index, best_index+1]]
            global_best_param_val, global_best_loglkl = self.get_bestfit_parabola(x, f)
            
        for param_val, bestfit in bestfit_dict.items():
            for idx_rep in range(self.config.profile.repetitions):
                if idx_rep in bestfit:
                    bestfit[idx_rep]['Delta_chi2'] = 2*(bestfit[idx_rep]['loglkl'] - global_best_loglkl)
        
        Delta_chi2 = []
        for param_val, bestfit in bestfit_dict.items():
            loglkl = bestfit[bestfit['best_rep']]['loglkl']
            if loglkl < np.inf:
                Delta_chi2.append(bestfit[bestfit['best_rep']]['Delta_chi2'])
            else:
                Delta_chi2.append(2*(bestfit['initial_loglkl'] - global_best_loglkl))
        Delta_chi2 = np.array(Delta_chi2)[ordering]

        profile = {
            'global_best_loglkl': global_best_loglkl,
            'global_best_param_val': global_best_param_val,
            'param_vals': param_vals,
            'loglkls': loglkls,
            'Delta_chi2': Delta_chi2
        }
        return profile

    def write_results(self, bestfit_dict, kernel) -> None:
        results_file = os.path.join(self.dir, f'{self.config.profile.parameter}.txt')
        param_names = [param_name for param_name in kernel.param['varying'] if param_name != self.config.profile.parameter]

        print(f"Writing current profile results to {results_file}")
        with open(results_file,"w") as file:
            # Write header
            file.write(f"{self.config.profile.parameter} \t Delta_chi2 \t -loglkl \t avg_best_loglkl \t initial_loglkl \t ")
            file.write(f" \t ".join([str(name) for name in param_names])+"\n")
            # Write results
            for fixed_val, bestfits_fixed_val in bestfit_dict.items():
                bestfit = bestfits_fixed_val[bestfits_fixed_val['best_rep']]
                if bestfit['loglkl'] == np.inf:
                    file.write(f"{fixed_val:.10f} \t --- no optimisations finished ---\n")
                else:
                    file.write(f"{fixed_val:.10f} \t {bestfit['Delta_chi2']:.10e} \t {bestfit['loglkl']:.10e} \t ")
                    file.write(f"{bestfits_fixed_val['avg_loglkl']:.10e} \t {bestfits_fixed_val['initial_loglkl']:.10e} \t ")
                    file.write(" \t ".join([str(np.round(bestfit['position'][name][0], 10)) for name in param_names])+"\n")

    def compute_intervals_neyman(self, profile):
        """
            Computes confidence intervals using the Neyman construction

        """
        intervals = {}

        def interp_onto(xdata, ydata, new_xdata, kind='linear'):
            return interp1d(xdata, ydata, fill_value="extrapolate", kind=kind)(new_xdata)
        param_cont = np.linspace(np.min(profile['param_vals']), np.max(profile['param_vals']), 10000)
        chi2_interped = interp_onto(profile['param_vals'], profile['Delta_chi2'], param_cont, kind='linear')
        idx_min = np.argmin(np.abs(param_cont - profile['global_best_param_val']))

        for cl in self.config.profile.confidence_levels:
            chi2 = stats.chi2.ppf(cl, 1) 
            intervals[cl] = [None, None]
            if idx_min > 0:
                lower_idx = np.argmin(np.abs(chi2_interped[0:idx_min] - chi2))
                intervals[cl][0] = param_cont[lower_idx]
            if idx_min < len(param_cont) - 1:
                upper_idx = idx_min + np.argmin(np.abs(chi2_interped[idx_min:-1] - chi2))
                intervals[cl][1] = param_cont[upper_idx]
        return intervals
    
    def write_intervals(self, profile, intervals, kernel) -> None:
        interval_file = os.path.join(self.dir, f'{self.config.profile.parameter}_intervals.txt')
        print(f"Writing current profile intervals to {interval_file}...")
        with open(interval_file, "w") as file:
            file.write('C.L. \t C.I. \t bestfit (+dist. to upper bound)(-dist. to lower bound)\n')
            for cl, interval in intervals.items():
                file.write(f"{cl*100} %: \t {interval} \t {profile['global_best_param_val']}")
                upper_dist, lower_dist = None, None
                if interval[1] is not None:
                    upper_dist = interval[1] - profile['global_best_param_val']
                if interval[0] is not None:
                    lower_dist = profile['global_best_param_val'] - interval[0]
                file.write(f"(+{upper_dist})(-{lower_dist}) \n")
                

    def plot_profile(self, profile, intervals, bestfit_dict, kernel) -> None:
        param_vals = profile['param_vals']
        loglkls    = profile['loglkls']
        Delta_chi2 = profile['Delta_chi2']

        fig, ax = plt.subplots(1, 1, figsize=(self.fig['width'], self.fig['height']))
        ax.set_xlabel(self.config.profile.parameter, fontsize=self.fig['fontsize'])
        if self.config.profile.plot_Delta_chi2:
            ax.set_ylabel('Delta chi2', fontsize=self.fig['fontsize'])
            for idx, cl in enumerate(self.config.profile.confidence_levels):
                chi2 = stats.chi2.ppf(cl, 1)
                ax.plot(param_vals, chi2*np.ones(len(param_vals)), 'k', linestyle=self.fig['interval_styles'][idx], lw=0.5, label=f'{cl} C.L.')
                for bound in intervals[cl]:
                    if bound is not None:
                        ax.plot([bound, bound], [ax.get_ylim()[0], chi2], 'k', linestyle=self.fig['interval_styles'][idx], lw=0.7)
            ax.plot(param_vals, Delta_chi2, 'k.-', lw=0.9, label='current', zorder=2)
        else:
            ax.set_ylabel('-loglkl', fontsize=self.fig['fontsize'])
            ax.plot(param_vals, loglkls, 'k.-', lw=0.9, label='current', zorder=2)

        if self.config.profile.detailed_plot:
            # Plot initial profile
            param_vals, loglkls_initial = [], []
            for param_val, bestfit in bestfit_dict.items():
                param_vals.append(param_val)
                loglkls_initial.append(bestfit['initial_loglkl'])
            ordering = np.argsort(param_vals)
            param_vals = np.array(param_vals)[ordering]
            loglkls_initial = np.array(loglkls_initial)[ordering]
            if self.config.profile.plot_Delta_chi2:
                ax.plot(param_vals, 2*(loglkls_initial - profile['global_best_loglkl']), 'y.--', ms=5, lw=0.7, alpha=0.8, label='initial')
            else:
                ax.plot(param_vals, loglkls_initial, 'y.--', ms=5, lw=0.7, alpha=0.8, label='initial', zorder=3)
            
            # Plot points from each rep
            for param_val, bestfit in bestfit_dict.items():
                for idx_rep in range(self.config.profile.repetitions):
                    if idx_rep in bestfit: # and idx_rep != bestfit['best_rep']:
                        if self.config.profile.plot_Delta_chi2:
                            ax.plot(param_val, 2*(bestfit[idx_rep]['loglkl'] - profile['global_best_loglkl']), '.', alpha=0.5, ms=self.fig['rep_ms'][idx_rep], color=self.fig['rep_colors'][idx_rep])
                        else:
                            ax.plot(param_val, bestfit[idx_rep]['loglkl'], '.', alpha=0.5, ms=self.fig['rep_ms'][idx_rep], color=self.fig['rep_colors'][idx_rep])
            for idx_rep in range(self.config.profile.repetitions):
                ax.plot([], [], '.', alpha=0.5, ms=self.fig['rep_ms'][idx_rep], color=self.fig['rep_colors'][idx_rep], label=f"rep {idx_rep}")
            
            if self.config.kernel.type == 'analytical':
                # Compute profile with scipy for comparison
                fixed_param_range = [min(self.config.profile.values), max(self.config.profile.values)]
                fixed_param_cont  = np.linspace(*fixed_param_range, 20)
                kernel.set_fixed_parameters({self.config.profile.parameter: 0.0})
                scipy_profile = []
                for fixed_val in fixed_param_cont:
                    scipy_profile.append(kernel.get_scipy_profile(self.config.profile.parameter, fixed_val))
                if self.config.profile.plot_Delta_chi2:
                    ax.plot(fixed_param_cont, 2*(scipy_profile - profile['global_best_loglkl']), 'g--', lw=1.5, label='scipy')
                else:
                    ax.plot(fixed_param_cont, scipy_profile, 'g--', lw=1.5, label='scipy')
        
        #ax.legend(fontsize=self.fig['fontsize']*0.8, ncol=2)
        fig.tight_layout()
        fig.savefig(os.path.join(self.dir, f'{self.config.profile.parameter}.pdf'))
        plt.close()

    def get_bestfit_parabola(self, x_triple, f_triple):
        # pm, pj and pp are the points (ordered according to their first coordinates) within which the bestfit lies
        # interpolating polynomial taken from the implementation of https://github.com/mwt5345/class_ede/issues/1
        xm, xj, xp = x_triple[0], x_triple[1], x_triple[2]
        fm, fj, fp = f_triple[0], f_triple[1], f_triple[2]
        x_best = (fp*(xj - xm)*(xj + xm) + fj*(xm - xp)*(xm + xp) + fm*(-pow(xj,2) + pow(xp,2)))/(2.*(fp*(xj - xm) + fj*(xm - xp) + fm*(-xj + xp)))
        f_best = fm + pow(-(fp*pow(xj - xm,2)) + fj*pow(xm - xp,2) + fm*(xj - xp)*(xj - 2*xm + xp),2)/(4.*(xj - xm)*(xj - xp)*(xm - xp)*(fp*(-xj + xm) + fm*(xj - xp) + fj*(-xm + xp)))
        return [x_best, f_best]

    def plot_schedule(self, optimise_tasks):
        # Collect nested dict data structure
        iter_data = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(float))))
        for task in optimise_tasks:
            iter_data[task.optimise_settings['fixed_param_val']][task.optimise_settings['repetition_number']][task.optimise_settings['iteration_number']] = {
                'acceptance_rate': task.optimiser.bestfit['acceptance_rate'],
                'temperature': task.optimiser.settings['temperature'],
                'step_size': task.optimiser.settings['step_size'],
                'step_size_change': task.optimiser.settings['step_size_change'],
                'loglkl': task.optimiser.bestfit['loglkl']
            }
        plot_adaptive = self.config.profile.step_size_schedule == 'adaptive' and self.config.profile.step_size_adaptive_multiplier == 'adaptive'

        if plot_adaptive == 'adaptive':
            fig, ax = plt.subplots(4, 1, figsize=(self.fig['width'], 3*self.fig['height']))
        else:
            fig, ax = plt.subplots(3, 1, figsize=(self.fig['width'], 2.5*self.fig['height']))
        current_longest = {'len': 0}
        for fixed_val in self.config.profile.values:
            # Actually: Consider plotting the average over reps!
            for idx_rep in range(self.config.profile.repetitions):
                if fixed_val in iter_data:
                    if idx_rep in iter_data[fixed_val]:
                        iterlist = list(iter_data[fixed_val][idx_rep].keys())
                        order = np.argsort(iterlist)
                        iterlist = np.array(iterlist)[order]
                        loglkls = np.array([it['loglkl'] for it in iter_data[fixed_val][idx_rep].values()])[order]
                        acceptance_rates = np.array([it['acceptance_rate'] for it in iter_data[fixed_val][idx_rep].values()])[order]
                        step_sizes = np.array([it['step_size'] for it in iter_data[fixed_val][idx_rep].values()])[order]
                        ax[0].plot(iterlist, loglkls, '.-', alpha=0.8, ms=self.fig['rep_ms'][idx_rep])
                        ax[1].plot(iterlist, acceptance_rates, '.-', alpha=0.8, ms=self.fig['rep_ms'][idx_rep])
                        ax[2].plot(iterlist, step_sizes, '.-')    
                        if plot_adaptive == 'adaptive':
                            multipliers = np.array([it['step_size_change']['current_multiplier'] for it in iter_data[fixed_val][idx_rep].values()])[order]
                            ax[3].plot(iterlist, multipliers, '.-', alpha=0.8, ms=self.fig['rep_ms'][idx_rep])
                        if len(iterlist) > current_longest['len']:
                            current_longest = {
                                'len': len(iterlist),
                                'fixed_val': fixed_val,
                                'idx_rep': idx_rep
                            }
        iterlim = [-0.1, current_longest['len'] + 0.1]
        ax[0].set(ylabel='-loglkl', xlim=iterlim, xticks=[], yscale='log')
        ax[1].set(ylabel='acceptance rate', xlim=iterlim, xticks=[])
        ax[2].set(ylabel='step size & temp', xlim=iterlim, xticks=[], yscale='log')
        if plot_adaptive:
            ax[3].set(xlabel='iteration', ylabel='step multiplier', xlim=iterlim)
        if current_longest['len'] > 0:
            # Plot temperature of the run with most iterations
            iterlist = list(iter_data[current_longest['fixed_val']][current_longest['idx_rep']].keys())
            order = np.argsort(iterlist)
            iterlist = np.array(iterlist)[order]
            temperatures = np.array([it['temperature'] for it in iter_data[current_longest['fixed_val']][current_longest['idx_rep']].values()])[order]
            ax[2].plot(iterlist, temperatures, 'r.-', label='T')    
            handles, labels = ax[0].get_legend_handles_labels()
            if handles and labels:
                labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
                ax[0].legend(handles, labels, fontsize=0.5*self.fig['fontsize'])
            ax[0].legend(handles, labels, fontsize=0.5*self.fig['fontsize'])
            ax[2].legend()

        fig.tight_layout()
        fig.savefig(os.path.join(self.dir, f'{self.config.profile.parameter}_schedule.pdf'))
        plt.close()


from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
import shutil
from time import time, perf_counter
from types import NoneType
from typing import Any
import numpy as np 
from prospect.input import InputArgument

import sys 

def initialise_mcmc(config_mcmc, kernel, **mcmc_args):
    if config_mcmc.algorithm == 'MetropolisHastings':
        sampler = MetropolisHastings(config_mcmc, kernel, **mcmc_args)
    else:
        raise ValueError('You have specified a non-existing MCMC type.')
    return sampler

"""
    Chain structure

"""

class Chain:
    mults: list[float]
    loglkls: list[float]
    positions: dict[str, np.ndarray] # Keys are parameter names, vals are arrays of positions traversed

    def __init__(self, initial_mults=None, initial_loglkls=None, initial_positions=None) -> None:
        self.mults, self.loglkls, self.positions = [], [], defaultdict(list)
        if initial_mults is not None:
            if initial_loglkls is not None:
                if initial_positions is not None:
                    self.mults += initial_mults
                    self.loglkls += initial_loglkls
                    for param_name, param_vec in initial_positions.items():
                        self.positions[param_name] += list(param_vec)

    def last_varying_position(self, list_of_varying_param_names):
        return {param_name: [self.positions[param_name][-1]] for param_name in list_of_varying_param_names}

    @property
    def last_position(self):
        out = {param: [self.positions[param][-1]] for param in list(self.positions.keys())}
        return out

    @property
    def N(self):
        if len(self.mults) == len(self.loglkls):
                for param_vec in self.positions.values():
                    if len(param_vec) != len(self.mults):
                        break
                else:
                    return len(self.mults)
        raise ValueError('Invalid dimensions in Chain: Non-compatible length of mults, loglkls and positions.')
    
    @property
    def data(self):
        """Returns a [N_points, 2+N_params] numpy array of chain contents in the format of a MontePython chain"""
        pos = np.array(list(self.positions.values()))
        output = np.concatenate([np.stack([self.mults, self.loglkls], axis=0), pos], axis=0).T
        return output

    def __getitem__(self, idx):
        # Return the position at idx
        return {param_name: [param_vec[idx]] for param_name, param_vec in self.positions.items()}
    
    def push_position(self, new_loglkl, new_position):
        self.mults.append(1)
        self.loglkls.append(new_loglkl)
        for param_name, param_vec in self.positions.items():
            param_vec.append(new_position[param_name][0])
    
    def from_indices(self, index_list):
        """Returns a sub-chain of self with the points of the indices given"""
        positions = {}
        for param_name, param_vec in self.positions.items():
            positions[param_name] = np.take(param_vec, index_list)
        return Chain(list(np.take(self.mults, index_list)),
                     list(np.take(self.loglkls, index_list)),
                     positions)

def collapse_chains(chain_list: list[Chain]) -> Chain:
    total_chain = Chain()
    for chain in chain_list:
        total_chain.mults += chain.mults # list addition
        total_chain.loglkls += chain.loglkls
        for param_name, param_vec in chain.positions.items():
            total_chain.positions[param_name] += list(param_vec)
    return total_chain

def compute_covariance_matrix(chain_list: list[Chain]):
    total_chain = collapse_chains(chain_list)
    # bias=True to normalize by N_points, like MP, instead of N_points-1
    # fweights are frequency weights, i.e. multiplicities
    return np.cov(total_chain.data[:, 2:], rowvar=False, bias=True, fweights=total_chain.mults)

"""
    MCMC Algorithms

"""

class BaseMCMC(ABC):
    def __init__(self, config_mcmc, kernel, **mcmc_args):
        super().__init__()
        self.config_mcmc = config_mcmc
        self.kernel = kernel

        if self.config_mcmc.start_from_covmat is not None:
            if type(self.config_mcmc.start_from_covmat) != str:
                self.covmat = self.config_mcmc.start_from_covmat
            else:
                self.covmat = self.kernel.read_covmat(self.config_mcmc.start_from_covmat)
        else:
            self.covmat = self.kernel.get_default_covmat()

        if 'chain' in mcmc_args:
            self.chain = mcmc_args['chain']
        else:
            if self.config_mcmc.start_from_position is not None:
                initial_position = self.kernel.read_initial_position(self.config_mcmc.start_from_position)
            else:
                initial_position = self.kernel.get_default_initial_position()
            self.chain = Chain([1], [self.kernel.loglkl(initial_position)], initial_position)

    @abstractmethod
    def step(self):
        pass

    def run_steps(self, steps_per_iteration: int):
        times = []
        for idx_step in range(steps_per_iteration):
            time_ini = perf_counter()
            self.step()
            time_final = perf_counter()
            times.append(time_final - time_ini)
        print(f"Finished running MCMC for {steps_per_iteration} steps. Average step time is {np.mean(times)} s.")
    
    def run_minutes(self, minutes: float):
        tic = time()
        seconds = minutes*60
        elapsed = 0
        times = []
        while elapsed < seconds:
            time_ini = perf_counter()
            self.step()
            time_final = perf_counter()
            times.append(time_final - time_ini)
            elapsed = time() - tic
        print(f"Finished running MCMC for {minutes} minutes. Average step time is {np.mean(times)} s.")
    
    def finalize(self):
        # Delete montepython log folders
        # shutil.rmtree(self.kernel.mp_dir)
        # Make sure kernel is not dumped when saving
        pass

class MetropolisHastings(BaseMCMC):
    def __init__(self, config_mcmc, kernel, **mcmc_args):
        super().__init__(config_mcmc, kernel, **mcmc_args)

    def step(self):
        prop = self.get_proposal()
        logprior_prop = self.kernel.logprior(prop)
        if logprior_prop == np.inf:
            acceptance = 0
        else:
            loglkl_prop = self.kernel.loglkl(prop)
            # Assume uniform priors => acc. prob. is likelihood ratio
            logpost_current = self.chain.loglkls[-1] + self.kernel.logprior(self.chain.last_position)
            logpost_prop = loglkl_prop + logprior_prop
            # Implicit minus sign on loglkl in our convention
            acceptance = np.exp((logpost_current - logpost_prop)/self.config_mcmc.temperature) 

        rand = np.random.uniform(low=0, high=1)
        if acceptance > rand:
            self.chain.push_position(loglkl_prop, prop)
        else:
            self.chain.mults[-1] += 1
        
    def get_proposal(self):
        # Important to remember fixed values!
        prop = {fixed_param_name: [param_dict['fixed_value']] for fixed_param_name, param_dict in self.kernel.param['fixed'].items()}
        current_means = np.array(list(self.chain.last_varying_position(self.kernel.varying_param_names).values())).flatten()
        # Note that we multiply the covmat by step_size^2 since step_size is a multiplier on the std deviations
        prop_list = np.random.multivariate_normal(current_means, self.covmat*self.config_mcmc.step_size**2)
        for idx, varying_param_name in enumerate(self.kernel.param['varying']):
            prop[varying_param_name] = [prop_list[idx]]
        return prop

"""
    Definition of user arguments related to mcmc.py
"""

@dataclass
class Arguments:
    class algorithm(InputArgument):
        val_type = str
        allowed_values = ['MetropolisHastings']
        default = 'MetropolisHastings'

    class N_chains(InputArgument):
        val_type = int
    
    class steps_per_iteration(InputArgument):
        val_type = int | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['steps_per_iteration'] is not None:
                assert config['minutes_per_iteration'] is None
    
    class minutes_per_iteration(InputArgument):
        val_type = int | float | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['minutes_per_iteration'] is not None:
                assert config['minutes_per_iteration'] > 0
                assert config['steps_per_iteration'] is None
    
    class convergence_Rm1(InputArgument):
        val_type = float
    
    class unpack_at_dump(InputArgument):
        val_type = bool
        default = True
    
    class temperature(InputArgument):
        """
            Only implemented on the acceptance criterion, so actual 
            loglkl values are independent of this and can always be compared.
        """
        val_type = float 
        def get_default(self, config_yaml: dict[str, Any]):
            return 1.0
    
    class step_size(InputArgument):
        # Multiplier on the proposal width
        val_type = float
        def get_default(self, config_yaml: dict[str, Any]):
                    return 1.0
    
    class analyse_automatically(InputArgument):
        allowed_values = [True, False, 'yes', 'y']
        def get_default(self, config_yaml: dict[str, Any]):
            return False
        def validate(self, config: dict[str, Any]) -> None:
            if not config['unpack_at_dump']:
                raise ValueError("Must unpack MCMC in order to analyse; please set 'unpack_at_dump' to True.")
    
    class start_from_covmat(InputArgument):
        # Specifies the parameter covmat 
        # Specific form depends on which kernel is being used
        # For analytical kernel: Nested lists of numbers defining the matrix
        # For MontePython: string pointing to a '.covmat' file
        # For cobaya: Don't enter!
        # Default value is extracted from the kernel
        val_type = str | list | np.ndarray | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
    
    class start_from_position(InputArgument):
        # Point to start from
        # Specific form depends on which kernel is being used
        # For analytical kernel: List of numbers
        # For MontePython: string pointing to a '.bestfit' file
        # For cobaya: Don't enter!
        # Default value is extracted from the kernel
        val_type = str | list | np.ndarray | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
    
    algorithm: algorithm
    N_chains: N_chains
    steps_per_iteration: steps_per_iteration
    minutes_per_iteration: minutes_per_iteration
    convergence_Rm1: convergence_Rm1
    unpack_at_dump: unpack_at_dump
    temperature: temperature
    step_size: step_size
    analyse_automatically: analyse_automatically
    start_from_covmat: start_from_covmat
    start_from_position: start_from_position



import time
from types import NoneType
from typing import Type
import numpy as np 
from getdist import loadMCSamples
from prospect.input import Configuration
from prospect.kernels.initialisation import initialise_kernel
from prospect.mcmc import Chain, collapse_chains, compute_covariance_matrix
from prospect.tasks.base_task import BaseTask
from prospect.tasks.optimise_task import OptimiseTask

class InitialiseOptimiserTask(BaseTask):
    """
        This task initialises a profile from a finished MCMC
        The MCMC must be made with the chosen kernel 
        (i.e. cannot start a MontePython-based profile from a cobaya chain)

    """
    priority = 80.0

    def __init__(self, config: Configuration, fixed_param_val: float | NoneType, repetition_number: int = 0):
        super().__init__(config)
        if self.config.run.jobtype == 'profile':
            self.fixed_param_val = fixed_param_val
        self.repetition_number = repetition_number

    def run(self, _):
        kernel = initialise_kernel(self.config.kernel, self.config.io.dir, self.id)
        if self.config.run.jobtype == 'profile':
            try:
                self.profile_param_idx = np.where(np.array(list(kernel.param['varying'].keys())) == self.config.profile.parameter)[0][0]
            except Exception as e:
                print("EXCEPTION:")
                print(f"The requested profile parameter {self.config.profile.parameter} is not recognized by the kernel. Check that your kernel parameter file has the parameter enabled.")
                print("Original exception:")
                raise e
            kernel.set_fixed_parameters({self.config.profile.parameter: self.fixed_param_val})

        if self.config.profile.start_from_position is not None and self.config.profile.start_from_covmat is not None:
            self.set_initial_position(kernel, kernel.read_initial_position(self.config.profile.start_from_position))
            self.set_covmat(kernel.read_covmat(self.config.profile.start_from_covmat))
            print(f"Read initial position from {self.config.profile.start_from_position}\n and initial covmat from {self.config.profile.start_from_covmat}.")

        elif self.config.profile.start_from_profile is not None:
            if self.config.profile.start_from_mcmc is not None:
                print("You have set 'start_from_profile' and 'start_from_mcmc'. Initialising position from profile and covmat from mcmc.")
                self.set_covmat(compute_covariance_matrix([self.get_binned_mcmc_reduced(kernel)]))
            elif self.config.profile.start_from_covmat is not None:
                print("You have set 'start_from_profile' and 'start_from_covmat'. Initialising position from profile and covmat from files.")
                self.set_covmat(kernel.read_covmat(self.config.profile.start_from_covmat))

            from prospect.profile import load_profile
            start_profile = load_profile(self.config.profile.start_from_profile, direct_txt=True)
            if self.config.run.jobtype == 'profile':
                idx_closest = np.argmin(np.abs(start_profile[self.config.profile.parameter] - self.fixed_param_val))
                print(f"Set initial bestfit for {self.config.profile.parameter}={self.fixed_param_val} from the point {self.config.profile.parameter}={start_profile[self.config.profile.parameter][idx_closest]} in {self.config.profile.start_from_profile}.")
            else:
                # Start from best point of the profile
                idx_closest = np.argmin(start_profile['-loglkl'])
                print(f"Set initial bestfit for from the best point in the profile {self.config.profile.start_from_profile}.")
            position = {param: [start_profile[param][idx_closest]] for param in kernel.param['varying'].keys()}
            self.set_initial_position(kernel, position)
        
        elif self.config.profile.start_from_mcmc is not None:
            binned_points = self.get_binned_mcmc_reduced(kernel)
            self.set_covmat(compute_covariance_matrix([binned_points]))

            # Get bestfit among these points
            bestfit_index = np.argmin(binned_points.loglkls)
            bf = binned_points[bestfit_index]
            if self.config.run.jobtype == 'profile' and self.config.profile.parameter in bf:
                del bf[self.config.profile.parameter]
                print(f"Constructed local bestfit and covariance matrix around {self.config.profile.parameter}={self.fixed_param_val}.")
            self.set_initial_position(kernel, bf)
        else:
            raise ValueError('InitialiseOptimiserTask could not construct an initial position and/or covariance matrix. Make sure to supply either "start_from_mcmc" or "start_from_position" and "start_from_covmat" in the profile section of the input.')

    def emit_tasks(self) -> list[Type[BaseTask]]:
        optimise_settings = {
            'current_best_loglkl': self.initial_loglkl,
            'initial_position': self.initial_bestfit,
            'covmat': self.initial_covmat,
            'temperature': self.config.profile.temperature_range[0],
            'temperature_change': get_temperature_change(self.config),
            'step_size': self.config.profile.step_size_range[0],
            'step_size_change': get_initial_step_size_change(self.config),
            'iteration_number': 0,
            'repetition_number': self.repetition_number
        }
        if self.config.run.jobtype == 'profile':
            optimise_settings['fixed_param_val'] = self.fixed_param_val
        return [OptimiseTask(self.config, optimise_settings)]
    
    def get_binned_mcmc_reduced(self, kernel):
        tic = time.perf_counter()
        initial_chain = load_mcmc(self.config.profile.start_from_mcmc, kernel, collapse=True)
        print(f"Loaded MCMC {self.config.profile.start_from_mcmc} in {time.perf_counter() - tic:.4} s")
        for param_name in kernel.param['varying']:
            if param_name not in initial_chain.positions:
                # Future: Just ignore this parameter and find a random starting point
                raise KeyError(f"Parameter '{param_name}' not present in the MCMC I'm starting from. Ending initialisation.")
        # Remove derived parameters from the input chain
        for param_name in kernel.param['derived']:
            if param_name in initial_chain.positions:
                del initial_chain.positions[param_name]
        # Remove unneeded parameters from MCMC
        unneeded_params = []
        for param_name in initial_chain.positions:
            if param_name not in kernel.param['varying']:
                if param_name not in kernel.param['derived']:
                    if param_name not in kernel.param['fixed']:
                        unneeded_params.append(param_name)
        for name in unneeded_params:
            del initial_chain.positions[name]

        if self.config.run.jobtype == 'profile':
            # Get start_bin_fraction fraction of total points closest to the param_val
            return get_fraction_of_points_in_bin(initial_chain, self.fixed_param_val, self.config.profile.parameter, self.config.profile.start_bin_fraction)
        else:
            return initial_chain.from_indices(range(len(initial_chain.mults)))

    def set_initial_position(self, kernel, position):
        self.initial_bestfit = position
        self.initial_loglkl = kernel.loglkl(self.initial_bestfit)

    def set_covmat(self, covmat):
        self.initial_covmat = covmat
        if self.config.run.jobtype == 'profile':
            self.initial_covmat = np.delete(self.initial_covmat, [self.profile_param_idx], 0) # remove row of profile param
            self.initial_covmat = np.delete(self.initial_covmat, [self.profile_param_idx], 1) # remove column of profile param 


def get_temperature_change(config):
    temp_change = 0.
    if config.profile.temperature_schedule == 'exponential':
        # The temperature change parameter is the rate, i.e. the number that
        # the temperature is multiplied by at each iteration
        temp_change = (config.profile.temperature_range[1]/config.profile.temperature_range[0])**(1/config.profile.max_iterations)
    else:
        raise ValueError('Invalid temperature schedule.')
    return temp_change

def get_initial_step_size_change(config):
    if config.profile.step_size_schedule == 'exponential':
        # Same logic as in `get_temperature_change()`
        step_size_change = (config.profile.step_size_range[1]/config.profile.step_size_range[0])**(1/config.profile.max_iterations)
    elif config.profile.step_size_schedule == 'adaptive':
        if config.profile.step_size_adaptive_multiplier == 'adaptive':
            # The change parameter is a dict of required information
            step_size_change = {}
        else:
            # Not used in this case 
            step_size_change = None
    else:
        raise ValueError('Invalid step size schedule.')
    return step_size_change

"""
    Functions for estimating profile likelihoods from MCMC chains


"""

def load_mcmc(mcmc_path: str, kernel, collapse=False) -> list[Chain]:
    """Loads the contents of an unpacked MCMC into a list of chains."""
    samples = loadMCSamples(mcmc_path)
    param_name_list = [param_name.name for param_name in samples.getParamNames().names]
    
    def create_prospect_chain(getdist_samples, param_name_list):
        prospect_chain = Chain()
        prospect_chain.mults = getdist_samples.weights
        prospect_chain.loglkls = getdist_samples.loglikes
        for idx_param, param_name in enumerate(param_name_list):
            prospect_chain.positions[param_name] = getdist_samples.samples[:, idx_param]
        return prospect_chain

    if collapse:
        return create_prospect_chain(samples, param_name_list)
    else:
        getdist_chains = samples.getSeparateChains()
        prospect_chains = []
        for getdist_chain in getdist_chains:
            prospect_chains.append(create_prospect_chain(getdist_chain, param_name_list))
        return prospect_chains

def get_points_in_bin(chain: Chain, param_val, param_name, N_points):
    """
        Returns a sub-chain of chain input consisting of the 'N_points' points 
        closest to the value 'param_val' of the parameter 'param_name'
    """
    # Sort points by their absolute distance to param_val
    distance_points = np.abs(np.array(chain.positions[param_name]) - param_val)
    ordered_indices = np.argsort(distance_points)
    # Take the first N_points points
    output_indices = ordered_indices[:N_points]
    return chain.from_indices(output_indices)

def get_fraction_of_points_in_bin(chain, param_val, param_name, fraction):
    N_points = int(np.ceil(fraction*chain.N))
    return get_points_in_bin(chain, param_val, param_name, N_points)

def estimate_bestfit(chains, param_val, param_idx, bin_fraction=0.1):
    binned_points = get_fraction_of_points_in_bin(chains, param_val, param_idx, bin_fraction)
    bestfit_index = np.argmin(binned_points[:, 1])
    return binned_points[bestfit_index]




import os 
import pickle
import shutil
import yaml 
import numpy as np 
from dataclasses import dataclass
from typing import Any
from prospect.input import InputArgument
from prospect.tasks.base_task import BaseTask

def read_user_input(arg: str) -> dict[str, Any]:
    if os.path.isfile(arg):
        resume = False
        config = yaml.full_load(open(arg, 'r'))
    elif os.path.isdir(arg):
        resume = True
        # arg is folder, get the .yaml stored inside
        config = yaml.full_load(open(f"{arg}/log.yaml", 'r'))
        # Force the output directory to be the user input
        if config['io']['dir']:
            if arg != config['io']['dir']:
                print(f"The command line directory supplied is different from the output directory in the log.yaml file.")
                print(f"Changing the output directory to the command line directory.")
                config['io']['dir'] = arg
    else:
        raise ValueError('Invalid arguments to PROSPECT. Give either a .yaml input file or the folder of a previous PROSPECT run.')
    return config, resume

def load_config(arg: str):
    # arg is folder, restart from that folder
    if not os.path.isfile(f"{arg}/config.pkl"):
        raise ValueError(f'Could not find {arg}/config.pkl.')
    else:
        with open(f"{arg}/config.pkl", "rb") as config_file:
            config = pickle.load(config_file)
        return config

def load_state(arg: str):
    # arg is folder, restart from that folder
    if not os.path.isfile(f"{arg}/state.pkl"):
        raise ValueError('No state.pkl found in your argument folder. Please provide either a folder with a PROSPECT state.pkl dump or an input.yaml file.')
    else:
        with open(f"{arg}/state.pkl", "rb") as state_file:
            state = pickle.load(state_file)
        max_id = max(set().union(*[state.ongoing, [task.id for task in state.queued], state.unready, state.done]))
        BaseTask.idx_count = max_id + 1
        return state

def prepare_run(config):
    BaseTask.idx_count = 1
    if config.io.write:
        # Setup job directory
        if os.path.isdir(config.io.dir):
            if config.io.overwrite_dir:
                shutil.rmtree(config.io.dir)
        os.makedirs(config.io.dir, exist_ok=False)
        # Create kernel directory
        os.makedirs(os.path.join(config.io.dir, config.run.jobtype))
        if config.kernel.type in ['montepython', 'cobaya', 'analytical']:
            os.makedirs(os.path.join(config.io.dir, config.kernel.type))
            # If MP: Save MP stuff
            # If cobaya: Save cobaya stuff
        # Save log.yaml
        print(f"Saving potentially modified config to {config.io.dir}/log.yaml")
        for module_name, module in config.config_dict.items():
            for param_name, param_value in module.items():
                # yaml cannot save numpy arrays, so convert them to lists before storing
                if isinstance(param_value, np.ndarray):
                    config.config_dict[module_name][param_name] = param_value.tolist()
        with open(os.path.join(config.io.dir, "log.yaml"), 'w') as outfile:
            yaml.dump(config.config_dict, outfile)

def unpack_mcmc(param_dict, output_dir, jobname, *chain_list) -> None:
    """Unpacks the PROSPECT state into a usual cobaya-like MCMC folder."""
    write_parameters(param_dict, output_dir, jobname)
    write_chains(output_dir, jobname, *chain_list)

def find(s, ch):
    """Utility for finding indices of string where character matches ch
       from https://stackoverflow.com/questions/11122291/how-to-find-char-in-string-and-get-all-the-indexes"""
    return [i for i, ltr in enumerate(s) if ltr == ch]

def remove_double_subscripts_latex(latex_names):
    """Reformats all double subscripts in latex names
       e.g. changes Omega_A_B_C to Omega_{A,B,C}"""
    for idx, name in enumerate(latex_names):
        underscore_indices = find(name, '_')
        if len(underscore_indices) > 1:
            # Double subscript
            latex_names[idx] = name[0:underscore_indices[0]] + '{' + name[underscore_indices[0]:].replace('_', ',') + '}'
    return latex_names

def write_parameters(param_dict, output_dir, jobname, latex_names=None) -> None:
    """Writes parameter names and ranges."""

    # If a name contains *, replace it by _ so as to be understood by GetDist
    reformatted_names = {}
    for name in param_dict.keys():
        new_name = name
        # Put all name replacements here
        if '*' in name:
            new_name = name.replace('*', '_')
        reformatted_names[name] = new_name
    # Doing it this way preserves the old ordering of the dict 
    param_dict = {reformatted_names[name]: param_dict[name] for name in param_dict.keys()}

    if latex_names == None:
        latex_names = remove_double_subscripts_latex(list(param_dict.keys()))

    with open(f"{output_dir}/{jobname}.paramnames", 'w') as file:
        for name, latex_name in zip(param_dict.keys(), latex_names):
            file.write(f"\n{name} {latex_name}")
    
    with open(f"{output_dir}/{jobname}.ranges", 'w') as file:
        for name, param_item in param_dict.items():
            write_str = f"\n{name}"
            for bound in param_item['range']:
                if bound is None:
                    write_str += " N"
                else:
                    write_str += f" {str(bound)}"
            file.write(write_str)

def write_chains(output_dir, jobname, *chain_list) -> None:
    """Stores chain contents to disk."""
    for idx_chain, chain in enumerate(chain_list):
        np.savetxt(f"{output_dir}/{jobname}_{idx_chain}.txt",  chain.data, delimiter=' ')


"""
    Definition of user arguments related to io.py

"""

@dataclass
class Arguments:
    class jobname(InputArgument):
        val_type = str
        default = None

    class write(InputArgument):
        val_type = bool
        default = False

    class dir(InputArgument):
        val_type = str
        default = None
        def validate(self, config):
            # Check that write = True
            pass
    
    class overwrite_dir(InputArgument):
        allowed_values = [True, False, 'yes', 'y']
        def get_default(self, config_yaml: dict[str, Any]):
            return False
        def validate(self, config):
            # Check that write = True
            pass
    
    class snapshot_interval(InputArgument):
        val_type = float
        # Dumps snapshot whenever task finishes
        def get_default(self, config_yaml: dict[str, Any]):
            return 1.0
        def validate(self, config):
            # Check that write = True
            pass
    
    jobname: jobname
    write: write
    dir: dir
    overwrite_dir: overwrite_dir
    snapshot_interval: snapshot_interval


import ast
import contextlib
import copy
import os
import shutil
import sys
from typing import Any
import numpy as np 
from prospect.kernels.base_kernel import BaseKernel

class InitialiseMontePython:
    """
        Singleton class that initialises MP on first init and returns
        that instance at every other initialisation, meaning
        MP is only ever initialised once
        (thus does not support MP inits with different input...)
    """
    def __new__(cls, config_kernel, output_folder, id):
        if not hasattr(cls, 'mp'):
            print("Initialising MontePython on process for the first time.")
            with contextlib.redirect_stdout(open(os.path.join(f"{output_folder}/montepython", f'id_{id}.out'), 'w')):
                with contextlib.redirect_stderr(open(os.path.join(f"{output_folder}/montepython", f'id_{id}.err'), 'w')):
                    cls.mp = cls.initialise_montepython(config_kernel, output_folder, id)
            #cls.mp = cls.initialise_montepython(config_kernel, output_folder, id)
        else:
            print("MontePython was already initialised, returning old instance.")
            if config_kernel.param != cls.mp['param']:
                raise ValueError(f"Tried to initialise MontePython with .param file {config_kernel.param}, which is different to {cls.mp['param']}, which it was firstly initialised with.")
            elif config_kernel.conf != cls.mp['conf']:
                raise ValueError(f"Tried to initialise MontePython with .conf file {config_kernel.conf}, which is different to {cls.mp['conf']}, which it was firstly initialised with.")
        return cls.mp
    
    def __reduce__(self) -> str | tuple[Any, ...]:
        return (None, None)

    def initialise_montepython(config_kernel, output_folder, id):
        mp = {}
        sys.path.append(f"{config_kernel.path}")
        try:
            from initialise import initialise as mp_initialise
            import sampler
            mp['compute_lkl'] = sampler.compute_lkl
            mp['get_covmat'] = sampler.get_covariance_matrix
            mp['read_args'] = sampler.read_args_from_bestfit
            from io_mp import CosmologicalModuleError
            mp['cosmo_soft_exception'] = CosmologicalModuleError
        except Exception:
            raise ImportError(f'Could not import MontePython modules. Is the path you gave, {config_kernel.path}, correctly pointing to the /montepython_public/montepython directory?')

        mp_dir = f"{output_folder}/montepython/id_{id}"
        mp_command_input = f'run -p {config_kernel.param} --conf {config_kernel.conf} -o {mp_dir} --chain-number 0'
        mp['param'], mp['conf'] = config_kernel.param, config_kernel.conf
        mp['cosmo'], mp['data'], mp['command_line'], _ = mp_initialise(mp_command_input)
        mp['err_dir'], mp['out_dir'] = f"{output_folder}/montepython/id_{id}.err", f"{output_folder}/montepython/id_{id}.out"
        return mp 

class MontePythonKernel(BaseKernel):
    def initialise(self, config_kernel, output_folder):
        self.mp = InitialiseMontePython(config_kernel, output_folder, self.id)
        self.dimension = len(list(self.mp['data'].mcmc_parameters.keys()))
        self.output_folder = output_folder
        self.config_kernel = config_kernel

        from classy import CosmoSevereError
        self.severe_exception = CosmoSevereError
        self.computation_exception = self.mp['cosmo_soft_exception']

    def set_parameter_dict(self):
        self.mp_to_prospect_name = {}
        for param_name, param_dict in self.varying_param_dict.items():
            prospect_param_name = self.format_param_name(param_name)
            self.param['varying'][prospect_param_name] = {'range': param_dict['prior'].prior_range}
            self.mp_to_prospect_name[param_name] = prospect_param_name
        for param_name in self.get_mp_param_names('derived'):
            prospect_param_name = self.format_param_name(param_name)
            self.param['derived'][prospect_param_name] = {}
            self.mp_to_prospect_name[param_name] = prospect_param_name
        self.prospect_to_mp_name = {val: key for key, val in self.mp_to_prospect_name.items()}

    def format_param_name(self, param_name):
        # Translate a parameter name from MontePython format to PROSPECT format
        prospect_param = param_name
        if '*' in param_name:
            prospect_param = param_name.replace('*', '')
        return prospect_param
    
    def save_config(self):
        shutil.copy(self.config_kernel.conf, os.path.join(self.output_folder, 'montepython/log.conf'))
        shutil.copy(self.config_kernel.param, os.path.join(self.output_folder, 'montepython/log.param'))

    def _loglkl(self, position):
        for param_name, param in self.param['varying'].items():
            self.mp['data'].mcmc_parameters[self.prospect_to_mp_name[param_name]]['current'] = position[param_name][0]
        for param_name, param in self.param['fixed'].items():
            self.mp['data'].mcmc_parameters[self.prospect_to_mp_name[param_name]]['current'] = param['fixed_value']
        self.mp['data'].update_cosmo_arguments()
        with contextlib.redirect_stdout(open(self.mp['out_dir'], 'a+')):
            with contextlib.redirect_stderr(open(self.mp['err_dir'], 'a+')):
                out = self.mp['compute_lkl'](self.mp['cosmo'], self.mp['data'])
        return -out

    def logprior(self, position):
        return self.log_uniform_prior(position)

    def read_initial_position(self, config_initial_position):
        # Reads a MontePython .bestfit file
        # If a parameter is not given in the .bestfit file, default value
        # for that parameter will be set
        with contextlib.redirect_stdout(open(self.mp['out_dir'], 'a+')):
            with contextlib.redirect_stderr(open(self.mp['err_dir'], 'a+')):
                self.mp['read_args'](self.mp['data'], config_initial_position)
        initial = {}
        for param_name, param in self.varying_param_dict.items():
            if 'last_accepted' in param:
                initial[self.mp_to_prospect_name[param_name]] = [param['last_accepted']]
            else:
                initial[self.mp_to_prospect_name[param_name]] = [param['initial'][0]]
        return initial

    def get_default_initial_position(self):
        # Default: Set initial values to mean values given in MP parameter file
        initial = {}
        for param_name, param in self.varying_param_dict.items():
            initial[param_name] = [param['initial'][0]]
        return initial

    def read_covmat(self, config_covmat):
        self.mp['command_line'].cov = config_covmat
        with contextlib.redirect_stdout(open(self.mp['out_dir'], 'a+')):
            with contextlib.redirect_stderr(open(self.mp['err_dir'], 'a+')):
                eigval, eigvec, covmat = self.mp['get_covmat'](self.mp['cosmo'], self.mp['data'], self.mp['command_line'])
        return covmat

    def get_default_covmat(self):
        # Extract default from MontePython
        with contextlib.redirect_stdout(open(self.mp['out_dir'], 'a+')):
            with contextlib.redirect_stderr(open(self.mp['err_dir'], 'a+')):
                eigval, eigvec, covmat = self.mp['get_covmat'](self.mp['cosmo'], self.mp['data'], self.mp['command_line'])
        return covmat

    def get_mp_param_names(self, type):
        return self.mp['data'].get_mcmc_parameters([type])

    @property
    def varying_param_dict(self):
        return {param_name: self.mp['data'].mcmc_parameters[param_name] for param_name in self.mp['data'].get_mcmc_parameters(['varying'])}



from abc import ABC, abstractmethod
from datetime import date
import time
from typing import Any
import traceback
from prospect.input import Configuration

class BaseTask(ABC):
    idx_count = -1
    priority: float
    required_task_ids: list[int]
    data: Any # Result from self.run() method is stored here
    success: bool

    time_launch: date = None
    time_emit: date = None
    time_finish: date = None
    emitted_by: int = None

    def __init__(self, config: Configuration, required_task_ids=[]):
        self.config = config
        self.id = BaseTask.idx_count
        if self.id < 0:
            raise ValueError('Tried initializing a task with negative id, probably because emit_tasks was called by a worker process.')
        for req in required_task_ids:
            if req >= self.id:
                raise ValueError(f'Cannot require non-existent task of ID {req}.')
            elif req < 0:
                raise ValueError('Required task ID is negative.')
        self.required_task_ids = required_task_ids
        BaseTask.idx_count += 1

    @abstractmethod
    def run(self) -> None:
        pass

    def emit_tasks(self):
        # Any required information must be stored in self.data during call to run()
        return []

    def run_return_self(self, *args):
        self.success = True
        self.error = None
        try:
            tic = time.perf_counter()
            self.run(*args)
            toc = time.perf_counter()
            print(f"Finished task of type {self.type} and id {self.id} in {toc - tic:.3} seconds")
            self.finalize()
            return self
        except Exception as e:
            self.success = False
            self.error = traceback.format_exc()
            self.finalize()
            return self

    def __lt__(self, other):
        # Largest numerical value of priority is greatest
        return self.priority > other.priority

    def finalize(self):
        pass

    @property
    def type(self) -> str:
        return self.__class__.__name__



from abc import ABC, abstractmethod
from dataclasses import dataclass
import os
import numpy as np 
from typing import Any, Type
from types import NoneType
from prospect.input import InputArgument

class BaseKernel(ABC):
    def __init__(self, config_kernel, task_id, output_folder=None):
        self.set_default_errors()
        self.id = task_id
        self.initialise(config_kernel, output_folder)

        self.param = {
            'varying': {},
            'fixed': {},
            'derived': {}
        }
        self.set_parameter_dict()
        self.save_config()
    
    class NullException(Exception):
        pass

    def set_default_errors(self):
        """
            Two error types exist: 
                Severe exceptions will stop the process
                Computation exceptions will assign -inf as the 
                likelihood value of the proposed point and then continue the run

            Derivatives of BaseKernel can define their own specific error types.
        """
        # By default, all exceptions are severe
        self.severe_exception = Exception 
        self.computation_exception = self.NullException

    @abstractmethod
    def initialise(self, kernel_param):
        pass

    def set_fixed_parameters(self, fixed_param_dict):
        for param_name, fixed_value in fixed_param_dict.items():
            # Move parameter to the fixed dict
            self.param['fixed'][param_name] = self.param['varying'][param_name]
            self.param['fixed'][param_name]['fixed_value'] = fixed_value
            del self.param['varying'][param_name]
    
    @property
    def varying_param_names(self):
        return list(self.param['varying'].keys())
    
    @abstractmethod
    def set_parameter_dict(self):
        """
            Must set self.param, a dict with keys that are parameter names
        """
        pass

    def save_config(self):
        """
        Saves kernel-specific data in the kernel subdir based on type
            - Analytical: Nothing saved
            - MontePython: .conf and .param
        """
        pass

    def loglkl(self, prop):
        for fixed_param, param in self.param['fixed'].items():
            prop[fixed_param] = [param['fixed_value']]
        try:
            return self._loglkl(prop)
        except self.computation_exception as e:
            print(f"Soft exception {e} occurred. Trying a new proposal.")
            return np.inf
        except self.severe_exception:
            raise ValueError("Severe exception occurred in likelihood computation. Stopping process.")
    
    def log_uniform_prior(self, position):
        if self.outside_of_prior_bound(position):
            return np.inf
        # Note: Different normalization than cobaya, same as MontePython
        return 0.0
    
    def outside_of_prior_bound(self, prop):
        outside = False
        for param_name, param_val in prop.items():
            if param_name in self.param['varying']:
                prior = self.param['varying'][param_name]['range']
                if prior[0] is not None:
                    if param_val[0] < prior[0]:
                        outside = True
                        break
                if prior[1] is not None:
                    if param_val[0] > prior[1]:
                        outside = True
                        break
        return outside

    @abstractmethod
    def _loglkl(self, position):
        # Should return -log(likelihood); note the minus sign!
        pass

    @abstractmethod
    def logprior(self, position):
        # Should return -log(prior); note the minus sign!
        pass

    def logpost(self, position):
        return self.loglkl(position)*self.logprior(position)

    @abstractmethod
    def get_default_initial_position(self):
        # Return default initial position
        pass

    @abstractmethod
    def read_initial_position(self, config_initial_position):
        # Return initial position given the user input
        pass

    @abstractmethod
    def get_default_covmat(self):
        # Return default (initial) covmat
        pass

    @abstractmethod
    def read_covmat(self, config_covmat):
        # Return (initial) covmat given user input
        pass

    def finalize(self):
        pass

"""
    Definition of user arguments related to base_kernel.py
"""

@dataclass
class Arguments:
    class type(InputArgument):
        val_type = str
        allowed_values = ['analytical', 'montepython', 'cobaya']
    
    class param(InputArgument):
        val_type = str # either path to .param/.yaml-file or dict if analytical type
        def validate(self, config: dict[str, Any]) -> None:
            if not os.path.isfile(os.path.join(os.getcwd(), config['param'])):
                raise ValueError(f"The file pointed to in the 'param' field of the 'kernel' input, which has the value {config['param']}, could not be found.")
    
    class conf(InputArgument):
        val_type = str
        def get_default(self, config_yaml: dict[str, Any]):
            if config_yaml['kernel']['type'] == 'analytical' or config_yaml['kernel']['type'] == 'cobaya':
                return ''
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['type'] == 'montepython':
                if not os.path.isfile(os.path.join(os.getcwd(), config['param'])):
                    raise ValueError(f"The file pointed to in the 'conf' field of the 'kernel' input, which has the value {config['conf']}, could not be found.")
    
    class path(InputArgument):
        # Only use if MontePython: Should point to the 'montepython' directory
        val_type = str
        def get_default(self, config_yaml: dict[str, Any]):
            if config_yaml['kernel']['type'] == 'analytical' or config_yaml['kernel']['type'] == 'cobaya':
                return ''
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['type'] == 'montepython':
                if not os.path.isdir(os.path.join(os.getcwd(), config['path'])):
                    raise ValueError(f"The directory pointed to in the 'path' field of the 'kernel' input, which has the value {config['path']}, could not be found.")

    class debug(InputArgument):
        # For cobaya: Debug mode, drops debug files in kernel subfolder
        val_type = bool
        def get_default(self, config_yaml: dict[str, Any]):
            return False

    type: type
    param: param
    conf: conf
    path: path
    debug: debug



from abc import ABC, abstractmethod
import numpy as np 
from prospect.mcmc import MetropolisHastings, Chain
from prospect.mcmc import Arguments as MCMCArguments

def initialise_optimiser(config, kernel, optimise_settings):
    if config.profile.optimiser == 'simulated annealing':
        optimiser = SimulatedAnnealing(config, kernel, optimise_settings)
    else:
        raise ValueError('You have specified a non-existing optimizer type.')
    return optimiser

class BaseOptimiser(ABC):
    def __init__(self, config, kernel, optimise_settings):
        self.kernel = kernel
        self.config = config
        self.settings = optimise_settings
        if 'initial_position' in optimise_settings and any(optimise_settings['initial_position']):
            self.initial_position = optimise_settings['initial_position']
        else:
            raise ValueError('No point to start the optimisation from.')
        if 'covmat' in optimise_settings:
            self.covmat = optimise_settings['covmat']
        else:
            raise ValueError('No covmat to start the optimisation from.')

    @abstractmethod
    def optimise():
        pass

    @abstractmethod
    def set_bestfit():
        pass

    @abstractmethod
    def get_next_iteration_settings():
        pass

class SimulatedAnnealing(BaseOptimiser):
    def __init__(self, config, kernel, optimise_settings):
        super().__init__(config, kernel, optimise_settings)

        config_mcmc = MCMCArguments(
            'Metropolis Hastings',
            1, # N_chains
            config.profile.steps_per_iteration,
            config.profile.minutes_per_iteration,
            0.0, # Target R-1; here we use different convergence criterion
            False, # unpack_at_dump
            self.settings['temperature'],
            self.settings['step_size'],
            False, # analyse_automatically
            self.covmat
        )
        mcmc_args = {
            'chain': Chain(
                            [1], 
                            [self.kernel.loglkl(self.initial_position)],
                            self.initial_position
                        )
        }
        self.mcmc = MetropolisHastings(config_mcmc, self.kernel, **mcmc_args)

    def optimise(self):
        if self.config.run.jobtype == 'profile':
            print(f"Running a simulated annealing iteration of parameter {self.config.profile.parameter}={self.settings['fixed_param_val']} with temperature {self.settings['temperature']}")
        else:
            print(f"Running a simulated annealing iteration with temperature {self.settings['temperature']}")
        if self.config.profile.steps_per_iteration is not None:
            self.mcmc.run_steps(self.config.profile.steps_per_iteration)
        elif self.config.profile.minutes_per_iteration is not None:
            self.mcmc.run_minutes(self.config.profile.minutes_per_iteration)
    
    def set_bestfit(self):
        bestfit_idx = np.argmin(self.mcmc.chain.loglkls)
        self.bestfit = {
            'loglkl': self.mcmc.chain.loglkls[bestfit_idx],
            'acceptance_rate': len(self.mcmc.chain.loglkls)/np.sum(self.mcmc.chain.mults),
            'accepted_steps': len(self.mcmc.chain.loglkls),
            'iteration_number': self.settings['iteration_number'],
            'position': {}
        }
        for param_name, param_vec in self.mcmc.chain.positions.items():
            self.bestfit['position'][param_name] = [param_vec[bestfit_idx]]
        #print(f"Optimiser set bestfit loglkl={self.bestfit['loglkl']}. Chain length is {len(self.mcmc.chain.loglkls)}.")
    
    def get_next_iteration_settings(self):
        if self.config.profile.temperature_schedule == 'exponential':
            new_temp = self.settings['temperature']*self.settings['temperature_change']
        else:
            raise ValueError('Invalid temperature schedule.')
        if self.config.profile.step_size_schedule == 'exponential':
            new_step_size = self.settings['step_size']*self.settings['step_size_change']
        elif self.config.profile.step_size_schedule == 'adaptive':
            if self.config.profile.step_size_adaptive_multiplier == 'adaptive':
                if not 'previous_multiplier' in self.settings['step_size_change']:
                    # First iteration, cannot use history to set multiplier
                    # Hardcoded initial multiplier
                    if self.bestfit['acceptance_rate'] > self.config.profile.step_size_adaptive_interval[1]:
                        initial_multiplier = 0.05
                    elif self.bestfit['acceptance_rate'] < self.config.profile.step_size_adaptive_interval[0]:
                        initial_multiplier = -0.05
                    new_step_size = self.settings['step_size']*(1 + initial_multiplier)
                    self.settings['step_size_change']['previous_multiplier'] = initial_multiplier
                    self.settings['step_size_change']['previous_acceptance_rate'] = self.bestfit['acceptance_rate']
                elif not 'current_multiplier' in self.settings['step_size_change']:
                    # Second iteration
                    if self.bestfit['acceptance_rate'] > self.config.profile.step_size_adaptive_interval[1]:
                        new_multiplier = 0.1
                    elif self.bestfit['acceptance_rate'] < self.config.profile.step_size_adaptive_interval[0]:
                        new_multiplier = -0.1
                    new_step_size = self.settings['step_size']*(1 + new_multiplier)
                    self.settings['step_size_change']['current_multiplier'] = new_multiplier
                    self.settings['step_size_change']['current_acceptance_rate'] = self.bestfit['acceptance_rate']
                else:
                    # At least third iteration, can now use estimated linear response 
                    # on AR to set optimal multiplier targetting the middle point of the desired AR interval
                    AR_desired = np.mean(self.config.profile.step_size_adaptive_interval)
                    AR_current = self.bestfit['acceptance_rate']
                    AR_previous = self.settings['step_size_change']['previous_acceptance_rate']
                    m_current = self.settings['step_size_change']['current_multiplier']
                    m_previous = self.settings['step_size_change']['previous_multiplier']
                    m_new = (AR_desired - AR_current)/(AR_current - AR_previous)*(m_current - m_previous) + m_current
                    #m_new = (AR_desired - AR_current)/(AR_current - AR_previous)*m_current/(1 + m_current)

                    # For stability
                    m_new = min(m_new, 1.0)
                    m_new = max(-0.9, m_new)
                    
                    new_step_size = self.settings['step_size']*(1 + m_new)
                    
                    # Update for next iteration
                    self.settings['step_size_change'] = {
                        'previous_acceptance_rate': AR_current,
                        'current_multiplier': m_new,
                        'previous_multiplier': m_current
                    }
            else:
                if self.bestfit['acceptance_rate'] > self.config.profile.step_size_adaptive_interval[1]:
                    # Acceptance rate is large, increase step size
                    new_step_size = self.settings['step_size']*(1 + self.config.profile.step_size_adaptive_multiplier)
                elif self.bestfit['acceptance_rate'] < self.config.profile.step_size_adaptive_interval[0]:
                    # Acceptance rate is small, reduce step size
                    new_step_size = self.settings['step_size']*(1 - self.config.profile.step_size_adaptive_multiplier)
                else:
                    # Step size is inside target interval, don't change
                    new_step_size = self.settings['step_size']
        else:
            raise ValueError('Invalid temperature schedule.')
        
        optimise_settings = {
            'current_best_loglkl': self.bestfit['loglkl'],
            'initial_position': self.mcmc.chain.last_position,
            'covmat': self.covmat,
            'temperature': new_temp,
            'temperature_change': self.settings['temperature_change'],
            'step_size': new_step_size,
            'step_size_change': self.settings['step_size_change']
        }
        if self.config.run.jobtype == 'profile':
            optimise_settings['fixed_param_val'] = self.settings['fixed_param_val']
        return optimise_settings



# Project
project  = 'PROSPECT'
author   = 'Emil Brinch Holm'
language = 'en'  # English

# Paths
html_static_path = ['static']
html_css_files = ['custom.css']
exclude_patterns = ['build']

# Extensions
extensions = [
    'sphinx_copybutton',
    'sphinx_rtd_theme',
    'sphinx_tabs.tabs',
]

# Never hide tab contents
sphinx_tabs_disable_tab_closing = True

# Theme
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    # General
    'display_version'           : False,
    'logo_only'                 : False,
    'prev_next_buttons_location': 'both',
    'style_external_links'      : True,
    # Navigation
    'collapse_navigation'        : False,
    'navigation_depth'           : -1,
    'sticky_navigation'          : True,
    'titles_only'                : False,
}

# Syntax highlighting of code blocks
import pygments.styles, pygments.token
def monkeypatch_pygments(name, base_name='default', attrs={}):
    import importlib, sys
    base_module = importlib.import_module('.'.join(['pygments', 'styles', base_name]))
    def name_to_class_name(name):
        return name.capitalize() + 'Style'
    base_class = getattr(base_module, name_to_class_name(base_name))
    attrs['styles'] = getattr(base_class, 'styles', {}) | attrs.pop('styles', {})
    class_name = name_to_class_name(name)
    Style = type(class_name, (base_class,), attrs)
    module = type(base_module)(name)
    setattr(module, class_name, Style)
    setattr(pygments.styles, name, module)
    pygments.styles.STYLE_MAP[name] = f'{name}::{class_name}'
    sys.modules['.'.join(['pygments', 'styles', name])] = module
pygments_style = 'concept'
monkeypatch_pygments(
    pygments_style,
    'friendly',
    {
        'background_color': '#f6f6f6',
        'styles': {
            pygments.token.Comment:       'italic #688F98',
            pygments.token.Name.Variable: '#d27a0a',
        },
    },
)

# Exclude Python and Bash prompts when copying code blocks
copybutton_prompt_text = r'>>> |\$ '
copybutton_prompt_is_regexp = True

# HTML
html_baseurl           = 'https://AarhusCosmology.github.io/prospect/'
html_context           = {'display_github': True}
# html_favicon           = html_static_path[0] + '/favicon.ico'
html_last_updated_fmt  = None
# html_logo              = html_static_path[0] + '/logo.png'
html_scaled_image_link = False
html_show_copyright    = False
html_show_sourcelink   = False
html_show_sphinx       = False
html_title             = 'PROSPECT Documentation'



from dataclasses import dataclass
import os
import sys
from types import NoneType
from typing import Any
from prospect.communication import Scheduler
from prospect.io import prepare_run, read_user_input, load_config, load_state
from prospect.input import Configuration, InputArgument

def master_process(config_yaml) -> bool:
    if config_yaml['run']['mode'] == 'mpi':
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank != 0:
            return False
    return True

def run(user_inp: str) -> None:
    config_yaml, resume = read_user_input(user_inp)
    if resume:
        config = load_config(user_inp)
        config.io.dir = config_yaml['io']['dir']
    else:
        config = Configuration(config_yaml)

    if config.run.mode == 'mpi':
        from mpi4py import MPI
        from mpi4py.futures import MPICommExecutor
        comm = MPI.COMM_WORLD
        if comm.Get_size() == 1:
            raise ValueError('You cannot run PROSPECT using MPI with only one process.')
        pool = MPICommExecutor
        pool_kwargs = {'comm': comm, 'root': 0}
    elif config.run.mode == 'threaded':
        from concurrent.futures import ProcessPoolExecutor
        pool = ProcessPoolExecutor
        pool_kwargs = {'max_workers': config.run.num_processes}
    elif config.run.mode == 'serial':
        from prospect.communication import SerialContext
        pool = SerialContext
        pool_kwargs = {}
    else:
        raise ValueError(f"Run mode '{config.run.mode}' not recognized. Choose either 'mpi', 'threaded' or 'serial'.")

    with pool(**pool_kwargs) as executor:
        if executor is not None:
            print(f"Running PROSPECT with mode *{config.run.mode}* on {config.run.num_processes} processes...")
            if resume:
                print(f"Resuming PROSPECT snapshop in {user_inp}.")
                state = load_state(user_inp)
            else:
                print(f"Starting PROSPECT from input file {user_inp}.")
                prepare_run(config)
                state = False
            scheduler = Scheduler(config, state)
            scheduler.delegate(executor)
            scheduler.finalize(executor)

    if config.run.mode == 'mpi':
        comm.Barrier()
        MPI.Finalize()
    sys.exit(0)

def run_from_shell() -> None:
    """
        Wrapper for run() using a setuptools entry point
        Allows running from command-line with the command `prospect input/test.yaml`
        Only takes arguments using argparse

    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', nargs='+')
    args = parser.parse_args()
    run(args.input_file[0])

def load(prospect_folder) -> Scheduler:
    """
        Loads a scheduler with all the information content of the 
        PROSPECT snapshot in prospect_pkl. Intended for use in interactive sessions.

        prospect_folder: 
            A folder with a previous prospect run.
            Must contain at least a snapshot.pkl and log.yaml.
    
    """
    config = load_config(prospect_folder)
    config.io.dir = prospect_folder
    state = load_state(prospect_folder)
    return Scheduler(config, state)

def analyse(prospect_folder):
    snapshot = load(prospect_folder)
    if snapshot.config.run.jobtype == 'profile':
        print(f"Analysing {prospect_folder} as a profile likelihood run.")
        analysis_task = snapshot.get_analysis_task()
    if snapshot.config.run.jobtype == 'global_optimisation':
        print(f"Analysing {prospect_folder} as a global optimisation likelihood run.")
        analysis_task = snapshot.get_analysis_task()

    elif snapshot.config.run.jobtype == 'mcmc':
        print(f"Analysing {prospect_folder} as an MCMC run.")
        from prospect.tasks.mcmc_task import AnalyseMCMCTask
        # Pick out the unique task of largest id belonging to each chain
        final_tasks = {}
        for chain_id, mcmc_task in {task.mcmc_args['chain_id']: task for task in snapshot.tasks.done.values() if task.type == 'MCMCTask'}:
            if chain_id not in final_tasks or mcmc_task.id > final_tasks[chain_id].id:
                final_tasks[chain_id] = mcmc_task
        snapshot.config.mcmc.analyse_automatically = True
        analysis_task = AnalyseMCMCTask(snapshot.config, required_task_ids=[task.id for task in final_tasks.values()])
        
    else:
        raise KeyError(f'Jobtype {snapshot.config.run.jobtype} not understood.')
    
    analysis_task.run([task for task in snapshot.tasks.done.values() if task.id in analysis_task.required_task_ids])
    print(f"\nFinished analysing {prospect_folder}. Enjoy!")

def analyse_from_shell():
    """
        Analyses the prospect folder if it is a profile job.
    
    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('prospect_folder', nargs='+')
    args = parser.parse_args()
    analyse(args.prospect_folder[0])


"""
    Definition of user arguments related to run.py
"""

@dataclass
class Arguments:
    class jobtype(InputArgument):
        val_type = str
        allowed_values = ['mcmc', 'profile', 'global_optimisation']
    
    class mode(InputArgument):
        val_type = str
        default = 'threaded'
        allowed_values = ['mpi', 'threaded', 'serial']
    
    class num_processes(InputArgument):
        val_type = int
        def get_default(self, config_yaml):
            if config_yaml['run']['mode'] == 'mpi':
                from mpi4py import MPI
                return MPI.COMM_WORLD.Get_size()
            elif config_yaml['run']['mode'] == 'threaded':
                return os.cpu_count() or 1
            elif config_yaml['run']['mode'] == 'serial':
                return 1
    
    class numpy_random_seed(InputArgument):
        val_type = float | int | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['numpy_random_seed'] is not None:
                import numpy as np 
                np.random.seed(config['numpy_random_seed'])

    jobtype: jobtype
    mode: mode
    num_processes: num_processes
    numpy_random_seed: numpy_random_seed



from collections import defaultdict
from copy import deepcopy
import os 
import time
from typing import Type
import matplotlib
import matplotlib.pyplot as plt 
import numpy as np 
from scipy.interpolate import interp1d
from scipy import stats
from prospect.input import Configuration
from prospect.kernels.initialisation import initialise_kernel
from prospect.tasks.base_analyse_task import BaseAnalyseTask
from prospect.tasks.optimise_task import OptimiseTask

# Switch matplotlib backend to one that doesn't pop up figures
matplotlib.use('Agg')

class AnalyseGlobalOptimisationTask(BaseAnalyseTask):
    def run(self, tasks: list[OptimiseTask]): 
        if not tasks:
            print(f"No tasks to analyse. Ending AnalyseProfileTask.")
            return
        print("Running global optimisation analysis script...")
        kernel = initialise_kernel(self.config.kernel, self.config.io.dir, self.id)
        optimise_tasks, initialise_tasks = [], []
        for task in tasks:
            if task.type == 'OptimiseTask':
                optimise_tasks.append(task)
            elif task.type == 'InitialiseOptimiserTask':
                initialise_tasks.append(task)
        bestfits = self.sort_tasks(optimise_tasks, initialise_tasks)
        self.write_results(bestfits, kernel)
        if self.config.profile.plot_schedule:
            tic = time.perf_counter()
            self.plot_schedule(optimise_tasks)
            print(f"Plotted schedule in {time.perf_counter() - tic:.5} s")

    def sort_tasks(self, optimise_tasks, initialise_tasks):
        bestfits = {}
        for idx_rep in range(self.config.profile.repetitions):
            current_bestfit = {'loglkl': np.inf}
            for task in optimise_tasks:
                if task.optimise_settings['repetition_number'] == idx_rep:
                    if task.optimiser.bestfit['loglkl'] < current_bestfit['loglkl']:
                        current_bestfit = task.optimiser.bestfit 
            bestfits[idx_rep] = current_bestfit # store the best in each rep
            best_rep = list(bestfits.keys())[np.argmin([rep['loglkl'] for rep in bestfits.values()])]
            bestfits[idx_rep]['best_rep'] = best_rep

            for initialise_task in initialise_tasks:
                # All reps are all initialised to the same point
                bestfits[idx_rep]['initial'] = initialise_task.initial_bestfit
                bestfits[idx_rep]['initial_loglkl'] = initialise_task.initial_loglkl
                bestfits[idx_rep]['covmat'] = initialise_task.initial_covmat
                break
        return bestfits

    def write_results(self, bestfits, kernel) -> None:
        results_file = os.path.join(self.dir, f'{self.config.profile.parameter}.txt')
        param_names = [param_name for param_name in kernel.param['varying'] if param_name != self.config.profile.parameter]

        print(f"Writing current optimisation results to {results_file}")
        with open(results_file,"w") as file:
            # Write header
            file.write(f"Repetition number \t -loglkl \t initial_loglkl \t ")
            file.write(f" \t ".join([str(name) for name in param_names])+"\n")
            # Write results
            for idx_rep in range(self.config.profile.repetitions):
                if bestfits[idx_rep]['loglkl'] == np.inf:
                    file.write(f"{idx_rep} \t --- no optimisations finished ---\n")
                else:
                    file.write(f"{idx_rep} \t {bestfits[idx_rep]['loglkl']:.10e} \t ")
                    file.write(f"{bestfits[idx_rep]['initial_loglkl']:.10e} \t ")
                    file.write(" \t ".join([str(np.round(bestfits[idx_rep]['position'][name][0], 10)) for name in param_names])+"\n")

    def plot_schedule(self, optimise_tasks):
        # Collect nested dict data structure
        iter_data = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
        # Sort optimisetasks by their repetition and iteration numbers
        for task in optimise_tasks:
            iter_data[task.optimise_settings['repetition_number']][task.optimise_settings['iteration_number']] = {
                'acceptance_rate': task.optimiser.bestfit['acceptance_rate'],
                'temperature': task.optimiser.settings['temperature'],
                'step_size': task.optimiser.settings['step_size'],
                'step_size_change': task.optimiser.settings['step_size_change'],
                'loglkl': task.optimiser.bestfit['loglkl']
            }
        plot_adaptive = self.config.profile.step_size_schedule == 'adaptive' and self.config.profile.step_size_adaptive_multiplier == 'adaptive'
        if plot_adaptive == 'adaptive':
            fig, ax = plt.subplots(4, 1, figsize=(self.fig['width'], 3*self.fig['height']))
        else:
            fig, ax = plt.subplots(3, 1, figsize=(self.fig['width'], 2.5*self.fig['height']))
        current_longest = {'len': 0}

        for idx_rep in range(self.config.profile.repetitions):
            if idx_rep in iter_data:
                iterlist = list(iter_data[idx_rep].keys())
                order = np.argsort(iterlist)
                iterlist = np.array(iterlist)[order]
                loglkls = np.array([it['loglkl'] for it in iter_data[idx_rep].values()])[order]
                acceptance_rates = np.array([it['acceptance_rate'] for it in iter_data[idx_rep].values()])[order]
                step_sizes = np.array([it['step_size'] for it in iter_data[idx_rep].values()])[order]
                ax[0].plot(iterlist, loglkls, '.-', alpha=0.8, ms=self.fig['rep_ms'][idx_rep])
                ax[1].plot(iterlist, acceptance_rates, '.-', alpha=0.8, ms=self.fig['rep_ms'][idx_rep])
                ax[2].plot(iterlist, step_sizes, '.-')    
                if plot_adaptive == 'adaptive':
                    multipliers = np.array([it['step_size_change']['current_multiplier'] for it in iter_data[idx_rep].values()])[order]
                    ax[3].plot(iterlist, multipliers, '.-', alpha=0.8, ms=self.fig['rep_ms'][idx_rep])
                if len(iterlist) > current_longest['len']:
                    current_longest = {
                        'len': len(iterlist),
                        'idx_rep': idx_rep
                    }
        iterlim = [-0.1, current_longest['len'] + 0.1]
        ax[0].set(ylabel='-loglkl', xlim=iterlim, xticks=[], yscale='log')
        ax[1].set(ylabel='acceptance rate', xlim=iterlim, xticks=[])
        ax[2].set(ylabel='step size & temp', xlim=iterlim, xticks=[], yscale='log')
        if plot_adaptive:
            ax[3].set(xlabel='iteration', ylabel='step multiplier', xlim=iterlim)
        if current_longest['len'] > 0:
            # Plot temperature of the run with most iterations
            iterlist = list(iter_data[current_longest['idx_rep']].keys())
            order = np.argsort(iterlist)
            iterlist = np.array(iterlist)[order]
            temperatures = np.array([it['temperature'] for it in iter_data[current_longest['idx_rep']].values()])[order]
            ax[2].plot(iterlist, temperatures, 'r.-', label='T')    
            handles, labels = ax[0].get_legend_handles_labels()
            if handles and labels:
                labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
                ax[0].legend(handles, labels, fontsize=0.5*self.fig['fontsize'])
            ax[0].legend(handles, labels, fontsize=0.5*self.fig['fontsize'])
            ax[2].legend()
        fig.tight_layout()
        fig.savefig(os.path.join(self.dir, f'optimisation_schedule.pdf'))
        plt.close()


from typing import Type
from prospect.input import Configuration
from prospect.tasks.base_task import BaseTask
from prospect.tasks.initialise_optimiser_task import InitialiseOptimiserTask
from prospect.tasks.initialise_optimiser_task import InitialiseOptimiserTask

class InitialiseProfileTask(BaseTask):
    """
        Task which finds a sampling of the profile and emits optimizations at each point in the profile.

    """
    priority = 90.0

    def run(self, _):
        pass
    
    def emit_tasks(self) -> list[Type[BaseTask]]:
        task_list = []
        if self.config.run.jobtype == 'global_optimisation':
            # Do global optimisation
            for idx_rep in range(self.config.profile.repetitions):
                task_list.append(InitialiseOptimiserTask(self.config, None, idx_rep))
        else:
            for idx_rep in range(self.config.profile.repetitions):
                for param_val in self.config.profile.values:
                    task_list.append(InitialiseOptimiserTask(self.config, param_val, idx_rep))
        return task_list

def initialise_profile_tasks(config: Configuration) -> list[Type[BaseTask]]:
    return [InitialiseProfileTask(config)]


from copy import deepcopy
from dataclasses import dataclass
from types import NoneType
from typing import Any
import numpy as np 
import os
import yaml
from prospect.communication import Scheduler
from prospect.input import InputArgument, Configuration
from prospect.run import load, load_state, load_config
from prospect.tasks.base_task import BaseTask
from prospect.tasks.analyse_profile_task import AnalyseProfileTask
from prospect.tasks.initialise_optimiser_task import InitialiseOptimiserTask, get_initial_step_size_change, get_temperature_change
from prospect.tasks.optimise_task import OptimiseTask


def reanneal(yaml_schedule, prospect_folder, override_queue=False):
    """
        Adds new OptimiseTasks to the queue, according to the given yaml file, and resumes.
            - yaml_schedule should point to a yaml file whose contents in dict-form are 
              {'profile': {dict with schedule specifications}}

            - prospect_folder should be the path to an existing prospect run



        Currently, the only full schedule specification should contain:
            optimiser, temperature_schedule, temperature_range,
            step_size_schedule, step_size_range,
            minutes_per_iteration OR steps_per_iteration,
            max_iterations, repetitions. 
        In principle, any argument inside the profile module can be changed, but changing
        other arguments than the above is untested.
    
    """
    new_schedule = yaml.full_load(open(yaml_schedule, 'r'))
    if not 'profile' in new_schedule:
        raise KeyError("Reanneal yaml file must contain 'profile' as the first and only key.")
    
    snapshot = load(prospect_folder)
    new_config_dict = deepcopy(snapshot.config.config_dict)
    for key, val in new_schedule['profile'].items():
        new_config_dict['profile'][key] = val
    new_config = Configuration(new_config_dict)
    new_config.io.dir = os.path.join(os.getcwd(), prospect_folder)

    if snapshot.tasks.ongoing or snapshot.tasks.unready or snapshot.tasks.queued:
        if override_queue:
            # WARNING: Removes all non-finised tasks. Note: ongoing tasks should already
            # have been moved to the ready queue upon loading.
            print(f"WARNING: Reannealing with override=True. Deleting all unfinished tasks in snapshot {prospect_folder}.")
            snapshot.tasks.ongoing = {}
            snapshot.tasks.unready = {}
            snapshot.tasks.queued = []
        else:
            raise ValueError('Cannot reanneal a run that did not finish. Please resume the run before reannealing.')

    scheduler = Scheduler(new_config, snapshot.tasks)

    # Create a dict with keys of param_vals and values that hold the id of the best task and its minimum loglkl
    optimisetasks = [task for task in snapshot.tasks.done.values() if task.type == 'OptimiseTask']
    best_tasks = {}
    for task in optimisetasks:
        if task.optimise_settings['fixed_param_val'] not in best_tasks:
            best_tasks[task.optimise_settings['fixed_param_val']] = {'id': task.id, 'loglkl': task.optimiser.bestfit['loglkl']}
        elif best_tasks[task.optimise_settings['fixed_param_val']]['loglkl'] > task.optimiser.bestfit['loglkl']:
            best_tasks[task.optimise_settings['fixed_param_val']] = {'id': task.id, 'loglkl': task.optimiser.bestfit['loglkl']}
    
    # Push new tasks
    submit_count = 0
    for idx_rep in range(new_config.profile.repetitions):
        for param_val in new_config.profile.values:
            if param_val in best_tasks:
                # Start from the old bestfit
                optimise_settings = {
                    'current_best_loglkl': best_tasks[param_val]['loglkl'],
                    'fixed_param_val': param_val,
                    'initial_position': snapshot.tasks.done[best_tasks[param_val]['id']].optimiser.bestfit['position'],
                    'covmat': snapshot.tasks.done[best_tasks[param_val]['id']].optimiser.covmat,
                    'temperature': new_config.profile.temperature_range[0],
                    'temperature_change': get_temperature_change(new_config),
                    'step_size': new_config.profile.step_size_range[0],
                    'step_size_change': get_initial_step_size_change(new_config),
                    'iteration_number': snapshot.tasks.done[best_tasks[param_val]['id']].optimiser.settings['iteration_number'] + 1,
                    'repetition_number': idx_rep
                }
                scheduler.push_task(OptimiseTask(scheduler.config, optimise_settings))
                submit_count += 1
            else:
                # Launch new optimisation
                scheduler.push_task(InitialiseOptimiserTask(scheduler.config, param_val, idx_rep))
                submit_count += 1
    
    scheduler.dump_snapshot()
    scheduler.status_update()
    print(f"Added {submit_count} optimisation tasks to snapshot {prospect_folder} with the schedule from {yaml_schedule}.")
    print(f"You may now resume the snapshot to run the tasks.")

def reanneal_from_shell():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-y', metavar='YAML', type=str, help="Path to a .yaml with the new schedule.")
    parser.add_argument('-o', metavar='FOLDER', type=str, help="Path to a PROSPECT folder to reanneal.")
    parser.add_argument('--override', help="Delete any unfinished tasks in the PROSPECT snapshot.", action='store_true')

    args = parser.parse_args()
    reanneal(args.y, args.o, args.override)

def load_profile(prospect_folder, direct_txt=False):
    if direct_txt:
        # If direct_txt is True, prospect_folder must point to the .txt file containing the results directly
        data_file = prospect_folder
    else:
        config = load_config(prospect_folder)
        data_file = f'{prospect_folder}/profile/{config.profile.parameter}.txt'
    if not os.path.isfile(data_file):
        raise KeyError(f"Could not load profile likelihood results: '{data_file}' not found.")
    results = np.genfromtxt(data_file, delimiter="\t", dtype=str, skip_header=0, autostrip=True)
    res_dict = {}
    for idx, entry in enumerate(results[0]):
        res_dict[entry] = np.array([float(val.strip('[]')) for val in results[1:, idx]])
    return res_dict
    

"""
    Definition of user arguments related to mcmc.py
"""

@dataclass
class Arguments:
    class parameter(InputArgument):
        val_type = str | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            if config_yaml['run']['jobtype'] == 'global_optimisation':
                # None corresponds to global optimisation
                return None
            else:
                raise ValueError("You need to specify the 'parameter' input of the profile section.")
    
    class values(InputArgument):
        """Parameter values where the profile is evaluated"""
        val_type = list[float] | list[int] | np.ndarray | list
        def get_default(self, config_yaml: dict[str, Any]):
            if config_yaml['run']['jobtype'] == 'global_optimisation':
                return []
            else:
                raise ValueError("You need to specify the 'values' input of the profile section.")
    
    class optimiser(InputArgument):
        allowed_values = ['simulated annealing']
        def get_default(self, config_yaml: dict[str, Any]):
            return 'simulated annealing'

    class temperature_schedule(InputArgument):
        allowed_values = ['exponential']
        def get_default(self, config_yaml: dict[str, Any]):
            return 'exponential'
        def validate(self, config: dict[str, Any]) -> None:
            assert config['optimiser'] == 'simulated annealing'
    
    class temperature_range(InputArgument):
        val_type = list[float] | list[np.float64] | np.ndarray | list
    
    class step_size_schedule(InputArgument):
        allowed_values = ['exponential', 'adaptive']
        def get_default(self, config_yaml: dict[str, Any]):
            return 'adaptive'
        def validate(self, config: dict[str, Any]) -> None:
            assert config['optimiser'] == 'simulated annealing'
    
    class step_size_range(InputArgument):
        val_type = list[float] | list[np.float64] | np.ndarray | list | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            None
        def validate(self, config: dict[str, Any]) -> None:
            if config['step_size_schedule'] == 'exponential':
                if config['step_size_range'] is None:
                    raise ValueError("Input 'step_size_range' is set to None, but must be set when using the 'exponential' schedule type.")

    class step_size_adaptive_interval(InputArgument):
        # The acceptance rate interval which will be aimed for
        val_type = list[float] | list[np.float64] | np.ndarray | list | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return [0.05, 0.15]

    class step_size_adaptive_multiplier(InputArgument):
        """
            The step size multiplier rate r at each iteration in the adaptive annealing
            i.e. if AR > adaptive_interval:
                new_ar = old_ar*(1 + r)
            elif AR < adaptive_interval:
                new_ar = old_ar*(1 - r)

            If 'adaptive', it will use estimate the linear response on AR to find the best value
            at each iteration
        """
        val_type = float | np.float64 | str
        def get_default(self, config_yaml: dict[str, Any]):
            return 'adaptive'
        def validate(self, config: dict[str, Any]) -> None:
            if isinstance(config['step_size_adaptive_multiplier'], str):
                if not config['step_size_adaptive_multiplier'] == 'adaptive':
                    raise ValueError("The only allowed string input for 'step_size_adaptive_multiplier' is 'adaptive'.'")
    
    class step_size_adaptive_initial(InputArgument):
        # Initial step size when using adaptive step size
        val_type = float | np.float64
        def get_default(self, config_yaml: dict[str, Any]):
            return 1.0
        def validate(self, config: dict[str, Any]) -> None:
            if config['step_size_schedule'] == 'adaptive':
                # Set initial step size variable in the range parameter
                config['step_size_range'] = [config['step_size_adaptive_initial'], None]

    class max_iterations(InputArgument):
        val_type = int | float
    
    class repetitions(InputArgument):
        val_type = int | float
        def get_default(self, config_yaml: dict[str, Any]):
            return 1

    class start_from_mcmc(InputArgument):
        val_type = str | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
    
    class start_bin_fraction(InputArgument):
        """
           Fraction of points (wrt. total amount in MCMC)
           around the profile parameter values included in the 
           initial profile generation
    
        """
        val_type = float
        def get_default(self, config_yaml: dict[str, Any]):
            return 0.1

    class start_from_covmat(InputArgument):
        # Specifies the parameter covmat 
        # Specific form depends on which kernel is being used
        # For analytical kernel: Nested lists of numbers defining the matrix
        # For MontePython: string pointing to a '.covmat' file
        # For cobaya: Don't enter!
        # Default value is extracted from the kernel
        val_type = str | list | np.ndarray | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
    
    class start_from_position(InputArgument):
        # Point to start from
        # Specific form depends on which kernel is being used
        # For analytical kernel: List of numbers
        # For MontePython: string pointing to a '.bestfit' file
        # For cobaya: Don't enter!
        # Default value is extracted from the kernel
        val_type = str | list | np.ndarray | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
    
    class start_from_profile(InputArgument):
        # .txt file with content loadable by load_profile
        # If the variables match, this starts the profile from the old profile
        # such that for each fixed param value in the new profile, the closest
        # bestfit in the old profile is taken as the starting point
        # If inputting this, remember to also put either start_from_mcmc or 
        # start_from_covmat in order to also get a covmat
        val_type = str | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['start_from_profile'] is not None:
                assert config['start_from_mcmc'] is not None or config['start_from_covmat'] is not None
    
        
    class chi2_tolerance(InputArgument):
        """
            Used as convergence criterion for the iterative optimisers:
            If chi2 changed less than this since last iteration, stop.
        
        """
        val_type = float | int | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
    
    class detailed_plot(InputArgument):
        """
            If true, plots initial points and each rep in profile
            Also, if using analytical kernel, plots the analytical profile

        """
        val_type = bool
        def get_default(self, config_yaml: dict[str, Any]):
            return False
    
    class plot_Delta_chi2(InputArgument):
        val_type = bool
        def get_default(self, config_yaml: dict[str, Any]):
            return True
    
    class plot_profile(InputArgument):
        val_type = bool
        def get_default(self, config_yaml: dict[str, Any]):
            return True

    class plot_schedule(InputArgument):
        val_type = bool
        def get_default(self, config_yaml: dict[str, Any]):
            return False
    
    class confidence_levels(InputArgument):
        # Computes confidence intervals at these levels
        val_type = list
        def get_default(self, config_yaml: dict[str, Any]):
            return [0.68, 0.95]

    parameter: parameter
    values: values
    optimiser: optimiser
    temperature_schedule: temperature_schedule
    temperature_range: temperature_range
    step_size_schedule: step_size_schedule
    step_size_range: step_size_range
    step_size_adaptive_interval: step_size_adaptive_interval
    step_size_adaptive_multiplier: step_size_adaptive_multiplier
    step_size_adaptive_initial: step_size_adaptive_initial
    #from prospect.mcmc import Arguments as mcmc_args
    #steps_per_iteration: mcmc_args.steps_per_iteration
    #minutes_per_iteration: mcmc_args.minutes_per_iteration
    """NOTE: The above doesn't work, so we duplicate code temporarily."""
    class steps_per_iteration(InputArgument):
        val_type = int | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['steps_per_iteration'] is not None:
                assert config['minutes_per_iteration'] is None
    
    class minutes_per_iteration(InputArgument):
        val_type = int | float | NoneType
        def get_default(self, config_yaml: dict[str, Any]):
            return None
        def validate(self, config: dict[str, Any]) -> None:
            if config['minutes_per_iteration'] is not None:
                assert config['minutes_per_iteration'] > 0
                assert config['steps_per_iteration'] is None
    steps_per_iteration: steps_per_iteration
    minutes_per_iteration: minutes_per_iteration

    max_iterations: max_iterations
    repetitions: repetitions
    start_from_mcmc: start_from_mcmc
    start_bin_fraction: start_bin_fraction
    chi2_tolerance: chi2_tolerance
    plot_profile: plot_profile
    detailed_plot: detailed_plot
    plot_Delta_chi2: plot_Delta_chi2
    plot_schedule: plot_schedule
    confidence_levels: confidence_levels
    start_from_covmat: start_from_covmat
    start_from_position: start_from_position
    start_from_profile: start_from_profile
    





import os
from typing import Any
import yaml
import numpy as np 
from prospect.input import safe_eval
from prospect.kernels.base_kernel import BaseKernel

class InitialiseCobaya:
    """
        Singleton-type class that initialises cobaya on first init and returns
        that instance at every other initialisation, meaning
        cobaya is only ever initialised once
        (thus does not support inits with different input...)
    """

    def __new__(cls, config_kernel, output_folder, id):
        if not hasattr(cls, 'model'):
            print("Initialising cobaya on process for the first time.")
            cls.model, cls.sampler, cls.yaml = cls.initialise_cobaya(config_kernel, output_folder, id)
            cls.param = config_kernel.param
        else:
            print("cobaya was already initialised, returning old instance.")
            if config_kernel.param != cls.param:
                raise ValueError(f"Tried to initialise MontePython with .param file {config_kernel.param}, which is different to {cls.param}, which it was firstly initialised with.")
        return cls.model, cls.sampler, cls.yaml
    
    def __reduce__(self) -> str | tuple[Any, ...]:
        return (None, None)
    
    def initialise_cobaya(config_kernel, output_folder, id):
        try:
            os.environ['COBAYA_NOMPI'] = "False"
            from cobaya.model import get_model
            from cobaya.yaml import yaml_load_file
            from cobaya.sampler import get_sampler
            extra_cobaya_settings = {
                'output': os.path.join(output_folder, f'cobaya/{id}/{id}')
            }
            if config_kernel.debug:
                extra_cobaya_settings['debug'] = os.path.join(output_folder, f'cobaya/{id}.debug')
            else:
                extra_cobaya_settings['debug'] = 40

            cobaya_yaml = yaml_load_file(config_kernel.param)
            cobaya_yaml.update(extra_cobaya_settings)
            cobaya_model = get_model(cobaya_yaml)
            cobaya_sampler = get_sampler({'mcmc': ''}, cobaya_model)
            return cobaya_model, cobaya_sampler, cobaya_yaml
        except ImportError:
            raise ImportError('Could not import cobaya. Make sure you have cobaya installed!')

class CobayaKernel(BaseKernel):
    def initialise(self, config_kernel, output_folder):
        self.cobaya_model, self.cobaya_sampler, self.cobaya_yaml = InitialiseCobaya(config_kernel, output_folder, self.id)
        self.dir = os.path.join(output_folder, 'cobaya')
        self.save_config()

        """
        if 'theory' in self.cobaya_yaml:
            if 'class' in self.cobaya_yaml:
                from classy import CosmoSevereError, CosmoComputationError
                self.severe_exception = CosmoSevereError
                self.computation_exception = CosmoComputationError
                # In camb, all errors are severe exceptions
        """

    def _loglkl(self, position):
        # Have to write it like this to unpack the value lists of 'positions'
        return -self.cobaya_model.loglike({param_name: pos[0] for param_name, pos in position.items()})[0]
    
    def logprior(self, position):
        return -self.cobaya_model.logprior({param_name: pos[0] for param_name, pos in position.items()})

    def set_parameter_dict(self):
        for param_name, param_dict in self.cobaya_model.parameterization.sampled_params_info().items():
            self.param['varying'][param_name] = {'prior': param_dict['prior']}
            if 'min' in param_dict['prior'] and 'max' in param_dict['prior']:
                self.param['varying'][param_name]['range'] = [param_dict['prior']['min'], param_dict['prior']['max']]
            else:
                self.param['varying'][param_name]['range'] = [None, None]
        for param_name in self.cobaya_model.parameterization.derived_params():
            self.param['derived'][param_name] = {}

    def save_config(self):
        with open(os.path.join(self.dir, "log.yaml"), 'w') as log_yaml:
            yaml.dump(self.cobaya_yaml, log_yaml)

    def read_initial_position(self, config_initial_position):
        # Must point to the my_cobaya_run_minize.bestfit or my_cobaya_run_minimize.bestfit.txt
        if not config_initial_position.endswith('.txt'):
            config_initial_position += '.txt'
        initial = np.genfromtxt(config_initial_position, names=True)
        initial_data = np.array(safe_eval(str(initial)))
        initial_position = {}
        for idx, value in enumerate(initial_data):
            name = initial.dtype.names[idx]
            if name in self.param['varying']:
                initial_position[name] = [value]
        return initial_position
    
    def get_default_initial_position(self):
        return {param: [val] for param, val in zip(self.cobaya_sampler.current_point.sampled_params, self.cobaya_sampler.current_point.values)}

    def read_covmat(self, config_covmat):
        # Imitates CovmatSampler._load_covmat from cobaya source code
        return np.atleast_2d(np.loadtxt(config_covmat))

    def get_default_covmat(self):
        return self.cobaya_sampler.initial_proposal_covmat()[0]



import os
from typing import Type
from prospect.kernels.initialisation import initialise_kernel
from prospect.mcmc import initialise_mcmc
from prospect.tasks.base_task import BaseTask
from prospect.analysis import get_gelman_rubin, getdist_gelman_rubin, analyse_mcmc
from prospect.input import Configuration
from prospect.io import unpack_mcmc

class MCMCTask(BaseTask):
    priority = 25.0

    def __init__(self, config: Configuration, **mcmc_args):
        super().__init__(config)
        self.mcmc_args = mcmc_args

    def run(self, _) -> None:
        kernel = initialise_kernel(self.config.kernel, self.config.io.dir, self.id)
        self.mcmc = initialise_mcmc(self.config.mcmc, kernel, **self.mcmc_args)
        if self.config.mcmc.steps_per_iteration is not None:
            self.mcmc.run_steps(self.config.mcmc.steps_per_iteration)
        elif self.config.mcmc.minutes_per_iteration is not None:
            self.mcmc.run_minutes(self.config.mcmc.minutes_per_iteration)
    
    def finalize(self):
        del self.mcmc.kernel

class AnalyseMCMCTask(BaseTask):
    priority = 75.0

    def __init__(self, config: Configuration, required_task_ids: list[int]):
        super().__init__(config, required_task_ids)
        self.dir = os.path.join(config.io.dir, config.run.jobtype)

    def run(self, mcmc_tasks: list[MCMCTask]): 
        chains = [task.mcmc.chain for task in mcmc_tasks]
        kernel = initialise_kernel(self.config.kernel, self.config.io.dir, self.id)

        if self.config.mcmc.unpack_at_dump:
            unpack_mcmc(kernel.param['varying'], self.dir, self.config.io.jobname, *chains)
            if self.config.mcmc.analyse_automatically:
                analyse_mcmc(self.dir, self.config.io.jobname)

        # Discrepancy wrt. the GetDist Gelman-Rubin; mine is stricter. Use GetDist's!
        if self.config.mcmc.unpack_at_dump:
            # Problem with this: It only works when unpacking the MCMC!
            self.conv = getdist_gelman_rubin(os.path.join(self.dir, self.config.io.jobname))
        else:
            raise ValueError("Currently cannot compute R-1 without unpacking MCMCs. Please set 'unpack_at_dump' to True.")
            self.conv = get_gelman_rubin(chains)

        self.mcmc_tasks = mcmc_tasks
        self.initial_positions = [chain.last_position for chain in chains]

    def emit_tasks(self) -> list[Type[BaseTask]]:
        if self.conv < self.config.mcmc.convergence_Rm1:
            print(f"Max. R-1 of {self.conv:.3} is below requirement {self.config.mcmc.convergence_Rm1}, finishing MCMC.")
            return [] # possibly emit some new finalising task 
        else:
            print(f"Max. R-1 of {self.conv:.3} is above requirement {self.config.mcmc.convergence_Rm1}, continuing MCMC.")
            return continue_mcmc_tasks(self.config, self.mcmc_tasks)

def initialise_mcmc_tasks(config: Configuration, mcmc_id: int) -> list[Type[BaseTask]]:
    task_list = []
    for idx_chain in range(config.mcmc.N_chains):
            mcmc_args = {'mcmc_id': mcmc_id, 'chain_id': idx_chain}
            task_list.append(MCMCTask(config, **mcmc_args))
    task_list.append(AnalyseMCMCTask(config, [task.id for task in task_list]))
    return task_list

def continue_mcmc_tasks(config: Configuration, mcmc_tasks: list[MCMCTask]) -> list[Type[BaseTask]]:
    task_list = []
    for mcmc_task in mcmc_tasks:
        mcmc_args = {
            'mcmc_id': mcmc_task.mcmc_args['mcmc_id'],
            'chain_id': mcmc_task.mcmc_args['chain_id'],
            'chain': mcmc_task.mcmc.chain
        }
        task_list.append(MCMCTask(config, **mcmc_args))
    task_list.append(AnalyseMCMCTask(config, [task.id for task in task_list]))
    return task_list



import os
import numpy as np 
from getdist import loadMCSamples, plots
import contextlib
from prospect.mcmc import Chain

def get_gelman_rubin(chain_list: list[Chain]) -> float:
    """Returns the R-1 Gelman Rubin convergence statistic."""
    remove_burnin(chain_list)

    # Possibly split chains into subchains; although MP doesn't do this I don't think
    param_names = chain_list[0].positions.keys()
    N_chains = len(chain_list)
    Rm1 = []
    for param_name in param_names:
        total_mean = np.mean([np.mean(np.stack(chain.positions)[param_name]) for chain in chain_list])
        var_between, var_within = 0.0, 0.0
        for chain in chain_list:
            positions = np.stack(chain.positions[param_name])
            N_points = positions.shape[0]
            var_between += (np.mean(positions) - total_mean)**2/(N_chains - 1)
            var_within += N_points/(N_points - 1)/N_chains*np.mean((positions - np.mean(positions))**2)
        Rm1.append(var_between/var_within)
    # Not using square root; same convention as MontePython and CosmoMC
    return np.max(Rm1)

def getdist_gelman_rubin(chain_folder: str) -> float:
    # Suppress console output from GetDist
    with contextlib.redirect_stdout(None):
        with contextlib.redirect_stderr(None):
            samples = loadMCSamples(chain_folder)
            return samples.getGelmanRubin()

def remove_burnin(chain_list: list[Chain]) -> None:
    pass

def get_effective_sample_size():
    pass

def compute_param_covariance():
    pass

def analyse_mcmc(output_dir, jobname) -> None:
    """Makes triangle plots and intervals from MCMC chains"""
    analysis_dir = os.path.join(output_dir, 'analysis')
    if not os.path.isdir(analysis_dir):
        print(f"Directory {analysis_dir} not found, creating...")
        os.makedirs(analysis_dir)
    with contextlib.redirect_stdout(open(os.path.join(analysis_dir, "getdist.out"), 'w')):
        with contextlib.redirect_stderr(open(os.path.join(analysis_dir, "getdist.err"), 'w')):
            try:
                samples = loadMCSamples(os.path.join(output_dir, jobname))
                with open(os.path.join(analysis_dir, f"{jobname}.stats"), 'w') as f:
                    print(samples.getMargeStats(), file=f)
                    f.write(f"R-1: {getdist_gelman_rubin(os.path.join(output_dir, jobname))}")
            except Exception as e:
                raise ValueError(f'Could not analyse chain. Try increasing amount of steps or time per iteration. Original exception:\n{e}')
            g = plots.get_subplot_plotter()
            g.triangle_plot(samples, filled=True)
            g.export(os.path.join(output_dir, 'analysis', f"{jobname}_triangle.pdf"))
